"""
Tracks changes between deployments.
"""
from pprint import pprint
#TODO: remove? largely deprecated, use the deploy module instead

from burlap import common
from burlap.decorators import task, runs_once
from burlap import Satchel

class ManifestSatchel(Satchel):

    name = 'manifest'

    @task
    @runs_once
    def show_current(self, name):
        ret = self.get_current(name)
        print('Current manifest for %s:' % name)
        pprint(ret, indent=4)

    @task
    @runs_once
    def show_last(self, name):
        ret = self.get_last(name)
        print('Last manifest for %s:' % name)
        pprint(ret, indent=4)

    @task
    @runs_once
    def get_current(self, name):
        name = name.strip().lower()
        func = common.manifest_recorder[name]
        return func()

    @task
    @runs_once
    def get_last(self, name):
        from burlap.deploy import deploy as deploy_satchel # pylint: disable=import-outside-toplevel
        name = common.assert_valid_satchel(name)
        last_thumbprint = deploy_satchel.get_previous_thumbprint()
        if last_thumbprint:
            if name in last_thumbprint:
                return last_thumbprint.get(name, type(self.genv)())
        return type(self.genv)()

    @task
    @runs_once
    def changed(self, name):
        from burlap.deploy import deploy # pylint: disable=import-outside-toplevel
        name = name.strip().lower()
        if name not in common.manifest_recorder:
            print('No manifest recorder has been registered for component "%s"' % name)
        else:
            last_thumbprint = deploy.get_previous_thumbprint()
            if last_thumbprint:
                if name in last_thumbprint:
                    last_manifest = last_thumbprint[name]
                    current_manifest = common.manifest_recorder[name]()
                    if last_manifest == current_manifest:
                        print('No')
                        return False
                    print('Yes')
                    return True
                print('Yes, first deployment for this component.')
                return True
            print('Yes, first deployment.')
            return True

manifest = ManifestSatchel()



import os
import sys

from fabric.api import env

from burlap.shelf import Shelf
from burlap.context import set_cwd
from burlap.common import set_verbose, get_verbose
from burlap.deploy import deploy as deploy_satchel
from burlap import load_role_handler
from burlap.common import all_satchels
from burlap.tests.functional_tests.base import TestCase

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
BASE_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '../../..'))

class DeployTests(TestCase):

    #TODO:re-enable once deploy rewrite merged
    def _test_deploy(self):
        """
        Creates a multi-site Apache Django powered web server with a MySQL backend.
        """
        try:
            set_verbose(True)
            assert 'apache_specifics' in env

            print('all_satchels:', sorted(all_satchels.keys()))
            assert len(all_satchels.keys()) == 61
            print('env.host_string:', env.host_string)
            print('env.hosts:', env.hosts)
            print('env.user:', env.user)
            assert env.host_string
            assert env.user

            # Delete any old tmp files
            PROJECT_DIR = '/tmp/burlap_test'
            if os.path.exists(PROJECT_DIR):
                #shutil.rmtree(PROJECT_DIR)
                os.system('rm -Rf %s/*' % PROJECT_DIR)
            else:
                os.makedirs(PROJECT_DIR)

            # Create our test virtualenv.
            PYTHON_EXE = os.path.split(sys.executable)[-1]
            VIRTUALENV_DIR = os.path.join(PROJECT_DIR, '.env')
            BURLAP_DIR = os.path.abspath(os.path.join(BASE_DIR, 'burlap'))
            BURLAP_BIN = os.path.abspath(os.path.join(BASE_DIR, 'bin/burlap-admin.py'))
            SITE_PACKAGES = os.path.join(VIRTUALENV_DIR, 'lib/%s/site-packages' % PYTHON_EXE)

            # Initialize project.
            kwargs = dict(
                project_dir=PROJECT_DIR,
                burlap_bin=BURLAP_BIN,
            )
            print('Initializing project skeleton...')
            assert os.path.isdir(PROJECT_DIR)
            with set_cwd(PROJECT_DIR):

                _status, _output = self.bash('{burlap_bin} skel multitenant'.format(**kwargs))
                print('_status, _output:', _status, _output)
                assert not _status

                # Symlink burlap.
                _status, _output = self.bash('ln -s %s %s' % (BURLAP_DIR, SITE_PACKAGES))
                #assert not _status

                # Add production role.
                VIRTUALENV_ACTIVATE = '. %s/bin/activate' % VIRTUALENV_DIR
                kwargs = dict(
                    project_dir=PROJECT_DIR,
                    activate=VIRTUALENV_ACTIVATE,
                    burlap_bin=BURLAP_BIN,
                )
                assert os.path.isdir(PROJECT_DIR)

                _status, _output = self.bash('{burlap_bin} add-role prod'.format(**kwargs))
                assert not _status

                # Test logging in to VM.
                print('env.host_string:', env.host_string)
                print('env.user:', env.user)
                print('env.key_filename:', env.key_filename)
                env.ROLE = 'prod'
                prod_settings = Shelf(filename='%s/roles/{role}/settings.yaml' % PROJECT_DIR)
                prod_settings['hosts'] = [env.host_string]
                assert prod_settings['hosts'][0]
                #prod_settings['host_string']
                prod_settings['user'] = env.user
                prod_settings['key_filename'] = env.key_filename
                prod_settings['is_local'] = False
                prod_settings['app_name'] = 'multitenant'
                kwargs = dict(
                    project_dir=PROJECT_DIR,
                )
                self.bash('ls -lah .')
                assert os.path.isdir(PROJECT_DIR)

                print('Testing hello world...')
                _status, _output = self.bash('.env/bin/fab prod:verbose=1 shell:command="echo hello"'.format(**kwargs))
                print('_status, _output:', _status, _output)
                assert not _status
                _status, _output = self.bash('ls -lah .')
                print('_status, _output:', _status, _output)

                print('Testing ifconfig...')
                _status, _output = self.bash('.env/bin/fab prod:verbose=1 shell:command="ifconfig"'.format(**kwargs))
                print('_status, _output:', _status, _output)
                assert 'inet addr:127.0.0.1' in _output

                # Add services.
                services = prod_settings.get('services', [])
                services.extend([
                    'apache',
                    #'hostname',
                    'mysql',
                    'mysqlclient',
                    'ntpclient',
                    'packager',
                    'pip',
                    'sshnice',
                    'tarball',
                    'timezone',
                    'ubuntumultiverse',
                    'unattendedupgrades',
                    #'user',
                ])
                prod_settings.set('services', services)
                prod_settings.set('sites', {
                    'multitenant': {
                        'apache_domain_template': 'multitenant.test.com',
                        'apache_domain_with_sub_template': 'multitenant.test.com',
                        'apache_domain_without_sub_template': 'multitenant.test.com',
                        'apache_server_aliases_template': 'multitenant.test.com',
                        'apache_ssl': False,
                        'apache_auth_basic': False,
                        'apache_enforce_subdomain': False,
                    },
                })
                prod_settings.set('pip_requirements', 'pip-requirements.txt')
                prod_settings.set('packager_initial_upgrade', False)

                # Confirm deployment changes are detected.
                #from burlap import role_prod as prod
                prod = load_role_handler('prod')
                prod()
                assert 'app_name' in env
                assert 'sites' in env
                env.host_string = env.hosts[0]

                print('-'*80)
                set_verbose(1)
                assert 'apache_specifics' in env
                print('Getting changed_components.verbose:', get_verbose())
                changed_components, deploy_funcs = deploy_satchel.get_component_funcs()
                changed_components = sorted(changed_components)
                expected_components = [
                    'APACHE',
                    'MYSQL',
                    'MYSQLCLIENT',
                    'NTPCLIENT',
                    'PACKAGER',
                    'PIP',
                    'SSHNICE',
                    'TARBALL',
                    'TIMEZONE',
                    'UBUNTUMULTIVERSE',
                    'UNATTENDEDUPGRADES',
                ]
                print('changed_components:', changed_components)
                print('expected_components:', expected_components)
                assert changed_components == expected_components
                deploy_funcs = sorted(deploy_funcs)
                print('deploy_funcs:', deploy_funcs)
                assert deploy_funcs == [
                    ('apache.configure', None),
                    ('mysql.configure', None),
                    ('mysqlclient.configure', None),
                    ('ntpclient.configure', None),
                    ('packager.configure', None),
                    ('pip.configure', None),
                    ('sshnice.configure', None),
                    ('tarball.configure', None),
                    ('timezone.configure', None),
                    ('ubuntumultiverse.configure', None),
                    ('unattendedupgrades.configure', None),
                ]

            # Deploy changes.

            # Confirm changes have been cleared.

            # Add Django site.

        finally:
            # Undo changes to the VM.
            pass



"""
Run test with:

    tox -c tox-full.ini -e py37-ubuntu_18_04_64 -- -s burlap/tests/test_vagrant_machines.py::TestVagrantMachines::test_machines_one

"""

from mock import patch

from burlap.tests.base import TestCase

class TestVagrantMachines(TestCase):

    def test_machines_one(self):
        with patch('burlap.vagrant.vagrant._status') as mock_status:
            mock_status.return_value = [('default', 'running')]
            from burlap.vagrant import vagrant # pylint: disable=import-outside-toplevel
            self.assertEqual(vagrant.machines(), ['default'])



import os
from pipes import quote
import random
import string

from fabric.api import hide

import six

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task


_SALT_CHARS = string.ascii_letters + string.digits + './'

CAT_KEY = 'cat-key'
UPLOAD_KEY = 'upload-key'

def _crypt_password(password):
    from crypt import crypt
    random.seed()
    salt = ''
    for _ in range(2):
        salt += random.choice(_SALT_CHARS)
    crypted_password = crypt(password, salt)
    return crypted_password


class UserSatchel(Satchel):

    name = 'user'

    def set_defaults(self):

        self.env.tmp_sudoers_fn = '/tmp/sudoers'

        self.env.key_type = 'rsa' # e.g. rsa|dsa
        self.env.key_bits = 2048 # e.g. 1024, 2048, or 4096
        self.env.key_filename_template = 'roles/{ROLE}/{host}-{username}.pem'
        self.env.key_perms = '600'

        self.env.home_template = '/home/{username}'
        #self.env.passwordless = True
        self.env.groups = {} # {username:[groups]}
        self.env.passwordless = {} # {username:True/False}
        self.env.passwords = {} # {username: password}
        self.env.reset_passwords_on_first_login = {} # {username: true/false}

        self.env.default_passwords = {} # {username:password}

        self.env.passwordless_method = CAT_KEY

    @task
    def enter_password_change(self, username=None, old_password=None):
        """
        Responds to a forced password change via `passwd` prompts due to password expiration.
        """
        from fabric.state import connections
        from fabric.network import disconnect_all
        r = self.local_renderer
#         print('self.genv.user:', self.genv.user)
#         print('self.env.passwords:', self.env.passwords)
        r.genv.user = r.genv.user or username
        r.pc('Changing password for user {user} via interactive prompts.')
        r.env.old_password = r.env.default_passwords[self.genv.user]
#         print('self.genv.user:', self.genv.user)
#         print('self.env.passwords:', self.env.passwords)
        r.env.new_password = self.env.passwords[self.genv.user]
        if old_password:
            r.env.old_password = old_password
        prompts = {
            '(current) UNIX password: ': r.env.old_password,
            'Enter new UNIX password: ': r.env.new_password,
            'Retype new UNIX password: ': r.env.new_password,
            #"Login password for '%s': " % r.genv.user: r.env.new_password,
#             "Login password for '%s': " % r.genv.user: r.env.old_password,
        }
        print('prompts:', prompts)

        r.env.password = r.env.old_password
        with self.settings(warn_only=True):
            ret = r._local("sshpass -p '{password}' ssh -o StrictHostKeyChecking=no {user}@{host_string} echo hello", capture=True)
            #code 1 = good password, but prompts needed
            #code 5 = bad password
            #code 6 = good password, but host public key is unknown
        if ret.return_code in (1, 6) or 'hello' in ret:
            # Login succeeded, so we haven't yet changed the password, so use the default password.
            self.genv.password = r.env.old_password
        elif self.genv.user in self.genv.user_passwords:
            # Otherwise, use the password or key set in the config.
            self.genv.password = r.env.new_password
        else:
            # Default password fails and there's no current password, so clear.
            self.genv.password = None
        print('using password:', self.genv.password)

        # Note, the correct current password should be set in host.initrole(), not here.
        #r.genv.password = r.env.new_password
        #r.genv.password = r.env.new_password
        with self.settings(prompts=prompts):
            ret = r._run('echo checking for expired password')
            print('ret:[%s]' % ret)
            do_disconnect = 'passwd: password updated successfully' in ret
            print('do_disconnect:', do_disconnect)
            if do_disconnect:
                # We need to disconnect to reset the session or else Linux will again prompt
                # us to change our password.
                disconnect_all()

                # Further logins should require the new password.
                self.genv.password = r.env.new_password

    @task
    def configure_keyless(self):
        self.generate_keys()
        self.passwordless()

    @task
    def togroups(self, user, groups):
        """
        Adds the user to the given list of groups.
        """

        r = self.local_renderer

        if isinstance(groups, six.string_types):
            groups = [_.strip() for _ in groups.split(',') if _.strip()]
        for group in groups:
            r.env.username = user
            r.env.group = group
            r.sudo('groupadd --force {group}')
            r.sudo('adduser {username} {group}')

    @task
    def passwordless(self, username, pubkey):
        """
        Configures the user to use an SSL key without a password.
        Assumes you've run generate_keys() first.
        """

        r = self.local_renderer

        r.env.username = username
        r.env.pubkey = pubkey
        if not self.dryrun:
            assert os.path.isfile(r.env.pubkey), \
                'Public key file "%s" does not exist.' % (str(r.env.pubkey),)

        first = os.path.splitext(r.env.pubkey)[0]
        r.env.pubkey = first+'.pub'
        r.env.pemkey = first+'.pem'
        r.env.home = r.env.home_template.format(username=username)

        # Upload the SSH key.
        r.sudo('mkdir -p {home}/.ssh')
        r.sudo('chown -R {user}:{user} {home}/.ssh')

        if r.env.passwordless_method == UPLOAD_KEY:
            put_remote_paths = self.put(local_path=r.env.pubkey)
            r.env.put_remote_path = put_remote_paths[0]
            r.sudo('cat {put_remote_path} >> {home}/.ssh/authorized_keys')
            r.sudo('rm -f {put_remote_path}')
        elif r.env.passwordless_method == CAT_KEY:
            r.env.password = r.env.default_passwords.get(r.env.username, r.genv.password)
            if r.env.password:
                r.local("cat {pubkey} | sshpass -p '{password}' ssh {user}@{host_string} 'cat >> {home}/.ssh/authorized_keys'")
            else:
                r.local("cat {pubkey} | ssh {user}@{host_string} 'cat >> {home}/.ssh/authorized_keys'")
        else:
            raise NotImplementedError

        # Disable password.
        r.sudo('cp /etc/sudoers {tmp_sudoers_fn}')
        r.sudo('echo "{username} ALL=(ALL) NOPASSWD: ALL" >> {tmp_sudoers_fn}')
        r.sudo('sudo EDITOR="cp {tmp_sudoers_fn}" visudo')

        r.sudo('service ssh reload')

        print('You should now be able to login with:')
        r.env.host_string = self.genv.host_string or (self.genv.hosts and self.genv.hosts[0])#self.genv.hostname_hostname
        r.comment('\tssh -i {pemkey} {username}@{host_string}')

    @task
    def generate_keys(self, username, hostname):
        """
        Generates *.pem and *.pub key files suitable for setting up passwordless SSH.
        """

        r = self.local_renderer

        #r.env.key_filename = r.env.key_filename or env.key_filename
        #assert r.env.key_filename, 'r.env.key_filename or env.key_filename must be set. e.g. roles/role/app_name-role.pem'
        r.env.key_filename = self.env.key_filename_template.format(
            ROLE=self.genv.ROLE,
            host=hostname,
            username=username,
        )
        if os.path.isfile(r.env.key_filename):
            r.pc('Key file {key_filename} already exists. Skipping generation.'.format(**r.env))
        else:
            r.local("ssh-keygen -t {key_type} -b {key_bits} -f {key_filename} -N ''")
            r.local('chmod {key_perms} {key_filename}')
            if r.env.key_filename.endswith('.pem'):
                src = r.env.key_filename+'.pub'
                dst = (r.env.key_filename+'.pub').replace('.pem', '')
#                 print('generate_keys:', src, dst)
                r.env.src = src
                r.env.dst = dst
                r.local('mv {src} {dst}')
        return r.env.key_filename

    @task
    def exists(self, name):
        """
        Check if a user exists.
        """
        with self.settings(hide('running', 'stdout', 'warnings'), warn_only=True):
            return self.run('getent passwd %s' % name).succeeded

    @task
    def create(self, username, groups=None, uid=None, create_home=None, system=False, password=None, home_dir=None):
        """
        Creates a user with the given username.
        """
        r = self.local_renderer
        r.env.username = username

        args = []

        if uid:
            args.append('-u %s' % uid)

        if create_home is None:
            create_home = not system

        if create_home is True:
            if home_dir:
                args.append('--home %s' % home_dir)
        elif create_home is False:
            args.append('--no-create-home')

        if password is None:
            pass
        elif password:
            crypted_password = _crypt_password(password)
            args.append('-p %s' % quote(crypted_password))
        else:
            args.append('--disabled-password')

        args.append('--gecos ""')

        if system:
            args.append('--system')

        r.env.args = ' '.join(args)
        r.env.groups = (groups or '').strip()
        r.sudo('adduser {args} {username} || true')
        if groups:
            for group in groups.split(' '):
                group = group.strip()
                if not group:
                    continue
                r.sudo('adduser %s %s || true' % (username, group))

    @task
    def expire_password(self, username):
        """
        Forces the user to change their password the next time they login.
        """
        r = self.local_renderer
        r.env.username = username
        r.sudo('chage -d 0 {username}')

    @task
    def configure(self):
        r = self.local_renderer

        lm = self.last_manifest
        lm_reset_passwords_on_first_login = lm.reset_passwords_on_first_login or {}
        lm_passwordless = lm.passwordless or {}
        lm_passwords = lm.passwords or {}
        lm_groups = lm.groups or {}

        # Make one-time password changes.
        just_changed = set()
        for username, ret in r.env.reset_passwords_on_first_login.items():
            if ret and not lm_reset_passwords_on_first_login.get(username):
                self.enter_password_change(username)
                just_changed.add(username)

        # Make passwordless logins.
        for username, is_passwordless in r.env.passwordless.items():
            if is_passwordless:
                if not lm_passwordless.get(username):
                    # If this user is passwordless, and we've not already created a passwordless
                    # login for them, then create one.
                    pubkey = self.generate_keys(username=username, host=self.genv.hostname_hostname)
                    self.passwordless(username=username, pubkey=pubkey)
            else:
                #TODO:expire old SSH key?
                pass

        # Update passwords.
        for username, password in r.env.passwords.items():
            if username in just_changed:
                continue
            if lm_passwords.get(username) != password:
                r.env.username = username
                r.env.password = password
                r.sudo('echo "{username}:{password}"|chpasswd')

        # Create new users.
        r = self.new_local_renderer
        new_users = set(r.env.groups).difference(lm_groups)
        for username in new_users:
            password = r.env.passwords.get(username, '')
            self.create(username=username, password=password)

        # Set groups.
        r = self.new_local_renderer
        for username, groups in r.env.groups.items():
            self.togroups(username, groups)

user = UserSatchel()



import socket
import sys
from io import BytesIO
from pprint import pprint
from functools import partial

import yaml

from fabric.api import execute, get

from burlap import ContainerSatchel
from burlap.constants import *
from burlap.decorators import task
from burlap.common import manifest_recorder, success_str, manifest_deployers_befores, topological_sort, resolve_deployer, \
    manifest_deployers_takes_diff, manifest_deployers, str_to_component_list, assert_valid_satchel, clean_service_name
from burlap import exceptions

def iter_dict_differences(a, b):
    """
    Returns a generator yielding all the keys that have values that differ between each dictionary.
    """
    common_keys = set(a).union(b)
    for k in common_keys:
        a_value = a.get(k)
        b_value = b.get(k)
        if a_value != b_value:
            yield k, (a_value, b_value)

def get_component_order(component_names):
    """
    Given a list of components, re-orders them according to inter-component dependencies so the most depended upon are first.
    """
    assert isinstance(component_names, (tuple, list))
    component_dependences = {}
    for _name in component_names:
        deps = set(manifest_deployers_befores.get(_name, []))
        deps = deps.intersection(component_names)
        component_dependences[_name] = deps
    component_order = list(topological_sort(component_dependences.items()))
    return component_order

def get_deploy_funcs(components, current_thumbprint, previous_thumbprint, preview=False):
    """
    Returns a generator yielding the named functions needed for a deployment.
    """
    for component in components:
        funcs = manifest_deployers.get(component, [])
        for func_name in funcs:

            #TODO:remove this after burlap.* naming prefix bug fixed
            if func_name.startswith('burlap.'):
                print('skipping %s' % func_name)
                continue

            takes_diff = manifest_deployers_takes_diff.get(func_name, False)

            func = resolve_deployer(func_name)
            current = current_thumbprint.get(component)
            last = previous_thumbprint.get(component)
            if takes_diff:
                yield func_name, partial(func, last=last, current=current)
            else:
                yield func_name, partial(func)

class DeploySatchel(ContainerSatchel):

    name = 'deploy'

    def set_defaults(self):
        self.env.lockfile_path = '~/burlap/deploy.lock'
        self.env.data_dir = '~/burlap'
        self._plan_funcs = None

    @task
    def init(self):
        """
        Initializes the configuration files on the remote server.
        """
        r = self.local_renderer
        #r.sudo('mkdir -p {data_dir}; chown {user}:{user} {data_dir}')
        r.run_or_local('mkdir -p {data_dir}')

    @task
    def purge(self):
        """
        The opposite of init(). Completely removes any manifest records from the remote host.
        """
        r = self.local_renderer
        r.run_or_local('[ -d {data_dir} ] && rm -Rf {data_dir} || true')

    @property
    def manifest_filename(self):
        """
        Returns the path to the manifest file.
        """
        r = self.local_renderer
        tp_fn = r.format(r.env.data_dir + '/manifest.yaml')
        return tp_fn

    def get_current_thumbprint(self, components=None):
        """
        Returns a dictionary representing the current configuration state.

        Args:
            components (str): Comma-separated list of components to deploy
        Returns:
            Thumbprint of the form:

            {
                component_name1: {key: value},
                component_name2: {key: value},
                ...
            }

        """
        components = str_to_component_list(components)
        if self.verbose:
            print('deploy.get_current_thumbprint.components:', components)
        manifest_data = {} # {component:data}
        for component_name, func in sorted(manifest_recorder.items()):
            self.vprint('Checking thumbprint for component %s...' % component_name)
            manifest_key = assert_valid_satchel(component_name)
            service_name = clean_service_name(component_name)
            if service_name not in self.genv.services:
                self.vprint('Skipping unused component:', component_name)
                continue
            if components and service_name not in components:
                self.vprint('Skipping non-matching component:', component_name)
                continue

            self.vprint('Retrieving manifest for %s...' % component_name)
            manifest_data[manifest_key] = func()
            if self.verbose:
                pprint(manifest_data[manifest_key], indent=4)

        return manifest_data

    def get_previous_thumbprint(self, components=None):
        """
        Returns a dictionary representing the previous configuration state.

        Args:
            components (str): Comma-separated list of components to deploy
        Returns:
            Thumbprint of the form:

            {
                component_name1: {key: value},
                component_name2: {key: value},
                ...
            }

        """
        components = str_to_component_list(components)
        tp_fn = self.manifest_filename
        if self.file_exists(tp_fn):
            fd = BytesIO()
            get(tp_fn, fd)
            tp_text = fd.getvalue()
            manifest_data = {}
            raw_data = yaml.load(tp_text, Loader=yaml.Loader)  # SafeLoader can't load native tuples
            for k, v in raw_data.items():
                manifest_key = assert_valid_satchel(k)
                service_name = clean_service_name(k)
                if components and service_name not in components:
                    continue
                manifest_data[manifest_key] = v
            return manifest_data

    @task
    def lock(self):
        """
        Marks the remote server as currently being deployed to.
        """
        self.init()
        r = self.local_renderer
        if self.file_exists(r.env.lockfile_path):
            raise exceptions.AbortDeployment('Lock file %s exists. Perhaps another deployment is currently underway?' % r.env.lockfile_path)
        self.vprint('Locking %s.' % r.env.lockfile_path)
        r.env.hostname = socket.gethostname()
        r.run_or_local('echo "{hostname}" > {lockfile_path}')

    @task
    def unlock(self):
        """
        Unmarks the remote server as currently being deployed to.
        """
        self.init()
        r = self.local_renderer
        if self.file_exists(r.env.lockfile_path):
            self.vprint('Unlocking %s.' % r.env.lockfile_path)
            r.run_or_local('rm -f {lockfile_path}')

    @task
    def fake(self, components=None):#, set_satchels=None):
        """
        Update the thumbprint on the remote server but execute no satchel configurators.

        components = A comma-delimited list of satchel names to limit the fake deployment to.
        set_satchels = A semi-colon delimited list of key-value pairs to set in satchels before recording a fake deployment.
        """

        self.init()

        # In cases where we only want to fake deployment of a specific satchel, then simply copy the last thumbprint and overwrite with a subset
        # of the current thumbprint filtered by our target components.
        if components:
            current_tp = self.get_previous_thumbprint() or {}
            current_tp.update(self.get_current_thumbprint(components=components) or {})
        else:
            current_tp = self.get_current_thumbprint(components=components) or {}

        tp_text = yaml.dump(current_tp)
        r = self.local_renderer
        r.upload_content(content=tp_text, fn=self.manifest_filename)

        # Ensure all cached manifests are cleared, so they reflect the newly deployed changes.
        self.reset_all_satchels()

    def get_component_funcs(self, components=None):
        """
        Calculates the components functions that need to be executed for a deployment.

        Args:
            components (str): Comma-separated list of components to deploy
        Returns:
            (component_order, plan_funcs)
        """

        current_tp = self.get_current_thumbprint(components=components) or {}
        previous_tp = self.get_previous_thumbprint(components=components) or {}

        if self.verbose:
            print('Current thumbprint:')
            pprint(current_tp, indent=4)
            print('Previous thumbprint:')
            pprint(previous_tp, indent=4)

        differences = list(iter_dict_differences(current_tp, previous_tp))
        if self.verbose:
            print('Differences:')
            pprint(differences, indent=4)
        component_order = get_component_order([k for k, (_, _) in differences])
        if self.verbose:
            print('component_order:')
            pprint(component_order, indent=4)
        plan_funcs = list(get_deploy_funcs(component_order, current_tp, previous_tp))

        return component_order, plan_funcs

    @task
    def preview(self, components=None, ask=0, show_issues=1):
        """
        Inspects differences between the last deployment and the current code state.

        Args:
            components (str): Comma-separated list of components to deploy
            ask (cast to int): Show confirmation prompt
            show_issues (cast to int): Show Jira issues linked to previewed commits
        """
        from burlap.git import gittracker, CURRENT_COMMIT
        from burlap.jirahelper import jirahelper

        ask = int(ask)
        show_issues = int(show_issues)

        self.init()

        component_order, plan_funcs = self.get_component_funcs(components=components)

        print('\n%i changes found for host %s.\n' % (len(component_order), self.genv.host_string))
        if component_order and plan_funcs:
            if self.verbose:
                print('These components have changed:\n')
                for component in sorted(component_order):
                    print((' '*4)+component)
            print('Deployment plan for host %s:\n' % self.genv.host_string)
            for func_name, _ in plan_funcs:
                print(success_str((' '*4)+func_name))
        if component_order:
            print()

        if show_issues and gittracker.is_selected:
            last_commit = gittracker.last_manifest.current_commit
            self.vprint('last_commit:', last_commit)
            current_commit = gittracker.current_manifest[CURRENT_COMMIT]
            self.vprint('current_commit:', current_commit)

            if last_commit and current_commit:
                # Show tickets between last deployment and now.
                tickets = jirahelper.get_tickets_between_commits(last_commit, current_commit)
                print('\nIssues to be deployed: {}\n'.format(', '.join(tickets) or 'None'))

        if ask and self.genv.host_string == self.genv.hosts[-1]:
            if component_order:
                if not input('Begin deployment? [yn] ').strip().lower().startswith('y'):
                    sys.exit(0)
            else:
                sys.exit(0)

    @task
    def push(self, components=None, yes=0, show_preview=0, show_issues=1):
        """
        Executes all satchel configurators to apply pending changes to the server.

        Args:
            components (str): Comma-separated list of components to deploy
            yes (cast to int): Bypass confirmation prompt
            show_preview (cast to int): Show deploy preview even if confirmation prompt is bypassed
            show_issues (cast to int): Show Jira issues linked to previewed commits
        """
        from burlap import notifier # pylint: disable=import-outside-toplevel

        service = self.get_satchel('service')
        yes = int(yes)
        show_preview = not yes or int(show_preview)
        show_issues = int(show_issues)

        try:
            self.lock() # Ensure that no one else can concurrently deploy to these hosts.

            if show_preview:
                # If we want to confirm the deployment with the user, and we're at the first server,
                # then run the preview.
                if self.genv.host_string == self.genv.hosts[0]:
                    execute(partial(self.preview, components=components, ask=not yes, show_issues=show_issues))

            notifier.notify_pre_deployment()
            component_order, plan_funcs = self.get_component_funcs(components=components)

            service.pre_deploy()
            for func_name, plan_func in plan_funcs:
                print('Executing %s...' % func_name)
                plan_func()
            self.fake(components=components)

            service.post_deploy()
            notifier.notify_post_deployment()

        except BaseException as exc:
            if getattr(exc, 'code', 1): # Don't send failure notifications for SystemExit with return code 0.
                notifier.notify_failed_deployment()
            raise

        finally:
            self.unlock()

deploy = DeploySatchel()



#!/usr/bin/env python
"""
Command line tool for initializing Burlap-structured Django projects.

e.g.

    burlap skel --name=myproject
"""
import argparse
import os
import sys

_path = os.path.normpath(os.path.join(os.path.realpath(__file__), '../..'))
sys.path.insert(0, _path)

import burlap # pylint: disable=wrong-import-position
from burlap.project import project # pylint: disable=wrong-import-position

ACTIONS = (
    SKEL,
    ADD_ROLE,
    CREATE_SATCHEL,
) = (
    'skel',
    'add-role',
    'create-satchel',
)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description=('Initialize and manage a Burlap structured project.'),
    )
    parser.add_argument(
        '-v', '--version',
        help='Show version.',
        action='version',
        version=burlap.__version__,
    )
    subparsers = parser.add_subparsers(dest='action')

    skel_parser = subparsers.add_parser(
        SKEL,
        help='Creates a skeleton project.')
    skel_parser.add_argument(
        'project_name',
        type=str,
        help='project name')
    skel_parser.add_argument(
        '--pip-requirements',
        type=str,
        default='Fabric',
        help='The default Python packages to install.')
    skel_parser.add_argument(
        '--virtualenv-dir',
        type=str,
        default='.env',
        help='The default virtualenv directory.')
    skel_parser.add_argument(
        '--roles',
        type=str,
        default='',#dev,prod',
        help='The default roles to populate.')
    skel_parser.add_argument(
        '--components',
        type=str,
        default='',
        help='The default components to enable (e.g. django).')
    skel_parser.add_argument(
        '--dj-version',
        type=str,
        default='',
        help='If the component dj is specified, the version of Django to install.')

    add_role_parser = subparsers.add_parser(
        ADD_ROLE,
        help='Adds a new role to the project.')
    add_role_parser.add_argument(
        'roles', metavar='roles', type=str, nargs='+',
        help='Names of roles to add.')

    create_satchel_parser = subparsers.add_parser(
        CREATE_SATCHEL,
        help='Adds a new role to the project.')
    create_satchel_parser.add_argument(
        'name', type=str,
        help='Names of the satchel to create.')

    args = parser.parse_args()
    if args.action == SKEL:
        project.create_skeleton(**args.__dict__)
    elif args.action == ADD_ROLE:
        project.add_roles(args.roles)
    elif args.action == CREATE_SATCHEL:
        project.create_satchel(args.name)
    else:
        raise NotImplementedError('Unknown action: %s' % (args.action))



"""
Debian packages
===============

This module provides tools to manage Debian/Ubuntu packages
and repositories.

"""
from fabric.api import hide, run, settings

import six

from burlap.utils import run_as_root
from burlap.files import file # pylint: disable=redefined-builtin

getmtime = file.getmtime
is_file = file.is_file


MANAGER = 'DEBIAN_FRONTEND=noninteractive apt-get'


def update_index(quiet=True):
    """
    Update APT package definitions.
    """
    options = "--quiet --quiet" if quiet else ""
    run_as_root("%s %s update" % (MANAGER, options))


def upgrade(safe=True):
    """
    Upgrade all packages.
    """
    manager = MANAGER
    if safe:
        cmd = 'upgrade'
    else:
        cmd = 'dist-upgrade'
    run_as_root("%(manager)s --assume-yes %(cmd)s" % locals(), pty=False)


def is_installed(pkg_name):
    """
    Check if a package is installed.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
        res = run("dpkg -s %(pkg_name)s" % locals())
        for line in res.splitlines():
            if line.startswith("Status: "):
                status = line[8:]
                if "installed" in status.split(' '):
                    return True
        return False


def install(packages, update=False, options=None, version=None):
    """
    Install one or more packages.

    If *update* is ``True``, the package definitions will be updated
    first, using :py:func:`~burlap.deb.update_index`.

    Extra *options* may be passed to ``apt-get`` if necessary.

    Example::

        import burlap

        # Update index, then install a single package
        burlap.deb.install('build-essential', update=True)

        # Install multiple packages
        burlap.deb.install([
            'python-dev',
            'libxml2-dev',
        ])

        # Install a specific version
        burlap.deb.install('emacs', version='23.3+1-1ubuntu9')

    """
    manager = MANAGER
    if update:
        update_index()
    if options is None:
        options = []
    if version is None:
        version = ''
    if version and not isinstance(packages, list):
        version = '=' + version
    if not isinstance(packages, six.string_types):
        packages = " ".join(packages)
    options.append("--quiet")
    options.append("--assume-yes")
    options = " ".join(options)
    cmd = '%(manager)s install %(options)s %(packages)s%(version)s' % locals()
    run_as_root(cmd, pty=False)


def uninstall(packages, purge=False, options=None):
    """
    Remove one or more packages.

    If *purge* is ``True``, the package configuration files will be
    removed from the system.

    Extra *options* may be passed to ``apt-get`` if necessary.
    """
    manager = MANAGER
    command = "purge" if purge else "remove"
    if options is None:
        options = []
    if not isinstance(packages, six.string_types):
        packages = " ".join(packages)
    options.append("--assume-yes")
    options = " ".join(options)
    cmd = '%(manager)s %(command)s %(options)s %(packages)s' % locals()
    run_as_root(cmd, pty=False)


def preseed_package(pkg_name, preseed):
    """
    Enable unattended package installation by preseeding ``debconf``
    parameters.

    Example::

        import burlap

        # Unattended install of Postfix mail server
        burlap.deb.preseed_package('postfix', {
            'postfix/main_mailer_type': ('select', 'Internet Site'),
            'postfix/mailname': ('string', 'example.com'),
            'postfix/destinations': ('string', 'example.com, localhost.localdomain, localhost'),
        })
        burlap.deb.install('postfix')

    """
    for q_name, _ in preseed.items():
        q_type, q_answer = _
        run_as_root('echo "%(pkg_name)s %(q_name)s %(q_type)s %(q_answer)s" | debconf-set-selections' % locals())


def get_selections():
    """
    Get the state of ``dkpg`` selections.

    Returns a dict with state => [packages].
    """
    with settings(hide('stdout')):
        res = run_as_root('dpkg --get-selections')
    selections = dict()
    for line in res.splitlines():
        package, status = line.split()
        selections.setdefault(status, list()).append(package)
    return selections


def apt_key_exists(keyid):
    """
    Check if the given key id exists in apt keyring.
    """

    # Command extracted from apt-key source
    gpg_cmd = 'gpg --ignore-time-conflict --no-options --no-default-keyring --keyring /etc/apt/trusted.gpg'

    with settings(hide('everything'), warn_only=True):
        res = run('%(gpg_cmd)s --fingerprint %(keyid)s' % locals())

    return res.succeeded


def _check_pgp_key(path, keyid):
    with settings(hide('everything')):
        return not run('gpg --with-colons %(path)s | cut -d: -f 5 | grep -q \'%(keyid)s$\'' % locals())


def add_apt_key(filename=None, url=None, keyid=None, keyserver='subkeys.pgp.net', update=False):
    """
    Trust packages signed with this public key.

    Example::

        import burlap

        # Varnish signing key from URL and verify fingerprint)
        burlap.deb.add_apt_key(keyid='C4DEFFEB', url='http://repo.varnish-cache.org/debian/GPG-key.txt')

        # Nginx signing key from default key server (subkeys.pgp.net)
        burlap.deb.add_apt_key(keyid='7BD9BF62')

        # From custom key server
        burlap.deb.add_apt_key(keyid='7BD9BF62', keyserver='keyserver.ubuntu.com')

        # From a file
        burlap.deb.add_apt_key(keyid='7BD9BF62', filename='nginx.asc'
    """

    if keyid is None:
        if filename is not None:
            run_as_root('apt-key add %(filename)s' % locals())
        elif url is not None:
            run_as_root('wget %(url)s -O - | apt-key add -' % locals())
        else:
            raise ValueError('Either filename, url or keyid must be provided as argument')
    else:
        if filename is not None:
            _check_pgp_key(filename, keyid)
            run_as_root('apt-key add %(filename)s' % locals())
        elif url is not None:
            tmp_key = '/tmp/tmp.burlap.key.%(keyid)s.key' % locals()
            run_as_root('wget %(url)s -O %(tmp_key)s' % locals())
            _check_pgp_key(tmp_key, keyid)
            run_as_root('apt-key add %(tmp_key)s' % locals())
        else:
            keyserver_opt = '--keyserver %(keyserver)s' % locals() if keyserver is not None else ''
            run_as_root('apt-key adv %(keyserver_opt)s --recv-keys %(keyid)s' % locals())

    if update:
        update_index()


def last_update_time():
    """
    Get the time of last APT index update

    This is the last modification time of ``/var/lib/apt/periodic/burlap-update-success-stamp``.

    Returns ``-1`` if there was no update before.

    Example::

        import burlap

        print(burlap.deb.last_update_time())
        # 1377603808.02

    """
    STAMP = '/var/lib/apt/periodic/burlap-update-success-stamp'
    if not is_file(STAMP):
        return -1
    return getmtime(STAMP)



import os

from burlap import common
from burlap.tests.base import TestCase

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

class DebugTests(TestCase):

    def setUp(self):
        super().setUp()

        common.set_verbose(True)
        # Ensure we're in burlap's root directory.
        os.chdir(os.path.abspath(os.path.join(CURRENT_DIR, '../..')))
        self._tmp_host_string = common.env.host_string
        self._tmp_is_local = common.env.is_local
        common.env.host_string = 'localhost'
        common.env.is_local = True

    def tearDown(self):
        common.env.host_string = self._tmp_host_string
        common.env.is_local = self._tmp_is_local

    def test_shell(self):
        from burlap.debug import debug # pylint: disable=import-outside-toplevel
        from burlap import shell # pylint: disable=import-outside-toplevel

        debug.verbose = True
        assert debug.genv.is_local
        ret = debug.shell(command="echo 'hello1'")
        print('ret:', ret)

        debug.genv.is_local = False
        ret = debug.shell(command="echo 'hello2'")
        print('ret:', ret)

        ret = shell(command="echo 'hello3'")
        print('ret:', ret)



import time
import logging
from pprint import pprint

import six

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task, runs_once
from burlap.common import print_success

GODADDY = 'godaddy'
BACKENDS = (
    GODADDY,
)

logger = logging.getLogger(__name__)

class DNSSatchel(Satchel):
    """
    Manages DNS zone records.
    """

    name = 'dns'

    def set_defaults(self):
        self.zones = []

    def update_dns_godaddy(self, domain, record_type, record):
        from godaddypy import Client, Account
        from godaddypy.client import BadResponse

        def get_domains(client):
            a = set()
            for d in client.get_domains():
                time.sleep(0.25)
                a.add(d)
            return a

        key = self.genv.godaddy_api_keys[domain]['key']
        secret = self.genv.godaddy_api_keys[domain]['secret']
        my_acct = Account(api_key=key, api_secret=secret)
        client = Client(my_acct)
        allowed_domains = get_domains(client)
        assert domain in allowed_domains, \
            'Domain %s is invalid this account. Only domains %s are allowed.' % (domain, ', '.join(sorted(allowed_domains)))
        logger.info('Adding record: %s %s %s', domain, record_type, record)
        if not self.dryrun:
            try:
                max_retries = 10
                for retry in six.moves.range(max_retries):
                    try:
                        client.add_record(
                            domain,
                            {
                                'data': record.get('ip', record.get('alias')),
                                'name': record['name'],
                                'ttl': record['ttl'],
                                'type': record_type.upper()
                            })
                        print_success('Record added!')
                        break
                    except ValueError as exc:
                        logger.error('Error adding DNS record on attempt %i of %i: %s', retry+1, max_retries, exc)
                        if retry + 1 == max_retries:
                            raise
                        time.sleep(3)
            except BadResponse as e:
                if e.message['code'] == 'DUPLICATE_RECORD':
                    logger.warning('Ignoring duplicate record.')
                else:
                    raise

    def get_last_zonefile(self, fn):
        lm = self.last_manifest
        zone_files = lm.zone_files or {}
        return zone_files.get(fn)

    @task
    @runs_once
    def update_dns(self, name=None):
        """
        Loop over zone file and add/update any missing entries.
        """
        from blockstack_zones import parse_zone_file

        r = self.local_renderer
        for zone_data in r.env.zones:
            zone_file = zone_data['file']
            domain = zone_data['domain']
            backend = zone_data['backend']
            types = zone_data['types']
            if backend not in BACKENDS:
                raise NotImplementedError('Unsupported backend: %s' % backend)
            logger.info('Processing zone file %s for domain %s.', zone_file, domain)
            zone_data = open(zone_file).read()
            zone_data = parse_zone_file(zone_data)
            if self.verbose:
                pprint(dict(zone_data), indent=4)

            #TODO:add differential update using get_last_zonefile()

            # Only update record types we're specifically in charge of managing.
            for record_type in types:
                record_type = record_type.lower()
                for record in zone_data.get(record_type):
                    backend_updater = getattr(self, 'update_dns_%s' % backend)
                    if name and name not in record['name']:
                        continue
                    backend_updater(domain=domain, record_type=record_type, record=record)

    def record_manifest(self):
        r = self.local_renderer
        manifest = super().record_manifest()
        manifest['zone_files'] = {}
        for zone_data in r.env.zones:
            zone_file = zone_data['file']
            manifest['zone_files'][zone_file] = open(zone_file).read()
        return manifest

    @task
    def configure(self):
        if self.genv.hosts and self.genv.host_string == self.genv.hosts[0]:
            self.update_dns()

dns = DNSSatchel()



import os
#import re
import shutil
import getpass
import traceback

from fabric.contrib.files import append
from fabric.api import settings
from fabric.exceptions import NetworkError

import burlap
from burlap.common import env
from burlap.tests.base import TestCase
from burlap.project import project
from burlap.context import set_cwd
from burlap.deploy import STORAGE_LOCAL

class DjTests(TestCase):

    def test_migrate(self):

        burlap_dir = os.path.abspath(os.path.split(burlap.__file__)[0])

        src_dir = '/tmp/test_dj_migrate/src'

        dj_version = [1, 10, 6]

        d = '/tmp/test_dj_migrate'
        if os.path.isdir(d):
            shutil.rmtree(d)
        os.makedirs(d)

        activate_cmd = '. {d}/.env/bin/activate;'.format(d=d)
        with set_cwd(d):
            project.create_skeleton(
                project_name='test_dj_migrate',
                roles='prod',
                components='dj',
                dj_version='.'.join(map(str, dj_version)),
            )

            assert not os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-1')
            assert not os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-2')

            # Simulate multiple remote hosts my creating aliases of localhost.
            # Note, for testing this on your localhost for a user without passwordless sudo,
            # you may have to run: `sudo chmod 777 /etc/hosts`
            # This won't work on Travis, where these will instead be set in .travis.yml.
            print('Modifying /etc/hosts...')
            env.host_string = 'localhost'
            env.hosts = [env.host_string]
            env.user = getpass.getuser()
            with settings(warn_only=True):
                for use_sudo in (False, True):
                    print('Trying with use_sudo:', use_sudo)
                    try:
                        ret = append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1', use_sudo=use_sudo)
                        ret = append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                        if ret is None:
                            break
                    except NetworkError:
                        print('Error modifying /etc/hosts using use_sudo=%s:' % use_sudo)
                        traceback.print_exc()

            os.system('ln -s %s %s/' % (burlap_dir, d))

            project.update_settings({
                    'plan_storage': STORAGE_LOCAL,
                    'plan_data_dir': os.path.join(d, 'plans'),
                    'services': ['dj'],
                    'dj_settings_module': 'test_dj_migrate.settings',
                    'default_site': 'testsite1',
                    'default_role': 'prod',
                    'dj_local_project_dir': 'src',
                    'dj_project_dir': '%s/src' % d,
                    'dj_manage_media': False,
                    'dj_manage_migrations': True,
                    'dj_manage_cmd': '%s/.env/bin/python manage.py' % d,
                    'dj_version': dj_version,
                    # This is necessary to stop get_current_hostname() from attempting to lookup our actual hostname.
                    '_ip_to_hostname': {
                        'test-dj-migrate-1': 'test-dj-migrate-1',
                        'test-dj-migrate-2': 'test-dj-migrate-2',
                    },
                },
                role='all')

            project.update_settings({
                    'hosts': ['test-dj-migrate-1', 'test-dj-migrate-2'],
                    'available_sites_by_host':{
                        'test-dj-migrate-1': [
                            'testsite1',
                        ],
                        'test-dj-migrate-2': [
                            'testsite2',
                        ]
                    },
                    'sites': {
                        'testsite1': {
                            'apache_domain_template': 'testsite1.test-dj-migrate-1.com',
                        },
                        'testsite2': {
                            'apache_domain_template': 'testsite2.test-dj-migrate-2.com',
                        },
                    },
                },
                role='prod')

            # Confirm both hosts are shown.
            kwargs = dict(
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} fab prod debug.list_hosts'.format(**kwargs))
            print('output:', output)
            assert 'test-dj-migrate-1' in output
            assert 'test-dj-migrate-2' in output

            #TODO:renable once deploy rewrite merged
            # Migrate built-in apps.
            #kwargs = dict(
                #activate_cmd=activate_cmd,
            #)
            #status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.run:yes=1,verbose=0'.format(**kwargs))
            #assert not status
            ## The migrations should have been run on both hosts.
            #assert '[test-dj-migrate-1] run:' in output
            #assert '[test-dj-migrate-2] run:' in output
            #assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-1')
            #assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/000/thumbprints/test-dj-migrate-2')

            ## Create custom app.
            #kwargs = dict(
                #src_dir='/tmp/test_dj_migrate/src',
                #activate_cmd=activate_cmd,
            #)
            #status, output = self.getstatusoutput('{activate_cmd} cd {src_dir}; python manage.py startapp myapp'.format(**kwargs))
            #assert os.path.isdir('/tmp/test_dj_migrate/src/myapp')

            ## Add myapp to installed apps list.
            #settings_fn = '/tmp/test_dj_migrate/src/test_dj_migrate/settings.py'
            #with open(settings_fn) as fin:
                #settings_text = fin.read()
            #p = re.compile(r'INSTALLED_APPS = \[([^\]]+)', flags=re.MULTILINE)
            #settings_text = p.sub(r"INSTALLED_APPS = [\1    'myapp',\n", settings_text)
            #with open(settings_fn, 'w') as fout:
                #fout.write(settings_text)
            #status, output = self.getstatusoutput('cd {src_dir}/test_dj_migrate; rm -f *.pyc'.format(**kwargs))
            #assert not status

            ## Confirm the correct version of Django was installed.
            #kwargs = dict(
                #src_dir=src_dir,
                #activate_cmd=activate_cmd,
            #)
            #status, output = self.getstatusoutput('{activate_cmd} cd {src_dir}; pip freeze | grep Django'.format(**kwargs))
            #assert '.'.join(map(str, dj_version)) in output

            ## Populate model.
            #open('/tmp/test_dj_migrate/src/myapp/models.py', 'w').write('''
#from django.db import models

#class MyModel(models.Model):
    #pass

#''')
            #kwargs = dict(
                #src_dir='/tmp/test_dj_migrate/src',
                #activate_cmd=activate_cmd,
            #)
            #status, output = self.getstatusoutput('{activate_cmd} cd {src_dir}; python manage.py makemigrations myapp'.format(**kwargs))
            #assert os.path.isfile('/tmp/test_dj_migrate/src/myapp/migrations/0001_initial.py')

            ## Apply migrations.
            #kwargs = dict(
                #activate_cmd=activate_cmd,
            #)
            ##cmd = '{activate_cmd} fab prod deploy.show_diff'.format(activate_cmd=activate_cmd)
            ##cmd = '{activate_cmd} fab prod deploy.run'.format(activate_cmd=activate_cmd)
            ##status, output = self.getstatusoutput('{activate_cmd} fab prod dj.configure:dryrun=1,verbose=1'.format(**kwargs))
            #status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.run:yes=1'.format(**kwargs))
            #print('output:', output)
            #assert not status
            ## The migrations should have been run on both hosts.
            #assert ('test-dj-migrate-1] run: export SITE=testsite1; export ROLE=prod; cd /tmp/test_dj_migrate/src; '
                #'/tmp/test_dj_migrate/.env/bin/python manage.py migrate') in output
            #assert ('test-dj-migrate-2] run: export SITE=testsite2; export ROLE=prod; cd /tmp/test_dj_migrate/src; '
                #'/tmp/test_dj_migrate/.env/bin/python manage.py migrate') in output

            #assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/001/thumbprints/test-dj-migrate-1')
            #assert os.path.isfile('/tmp/test_dj_migrate/plans/prod/001/thumbprints/test-dj-migrate-2')



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class ElasticSearchSatchel(ServiceSatchel):

    name = 'elasticsearch'

    def set_defaults(self):

        self.env.conf_path = '/etc/elasticsearch/elasticsearch.yml'
        self.env.script_engine_groovy_inline_search = False

        self.env.service_commands = {
            START:{
                UBUNTU: 'service elasticsearch start',
            },
            STOP:{
                UBUNTU: 'service elasticsearch stop',
            },
            DISABLE:{
                UBUNTU: 'chkconfig elasticsearch off',
                (UBUNTU, '14.04'): 'update-rc.d -f elasticsearch remove',
            },
            ENABLE:{
                UBUNTU: 'chkconfig elasticsearch on',
                (UBUNTU, '14.04'): 'update-rc.d elasticsearch defaults',
            },
            RESTART:{
                UBUNTU: 'service elasticsearch restart',
            },
            STATUS:{
                UBUNTU: 'service elasticsearch status',
            },
        }

    @property
    def packager_system_packages(self):
        return {
            DEBIAN: ['elasticsearch'],
            UBUNTU: ['elasticsearch'],
        }

    @task(precursors=['packager', 'user', 'cron'])
    def configure(self):
        r = self.local_renderer

        if self.env.enabled:

            if self.env.script_engine_groovy_inline_search:

                # Turn on online groovy search.
                r.sed(
                    filename=self.env.conf_path,
                    before='script.engine.groovy.inline.search: off',
                    after='script.engine.groovy.inline.search: on',
                    use_sudo=True,
                )
                # Remove off.
#                 r.sudo(
#                     "sed '/script.engine.groovy.inline.search: off/d' {conf_path}"\
#                         .format(conf_path=self.env.conf_path))
                # Add on.
#                 r.sudo(
#                     "echo 'script.engine.groovy.inline.search: on' >> {conf_path}"\
#                         .format(conf_path=self.env.conf_path))
#                 r.append(
#                     text='script.engine.groovy.inline.search: on',
#                     filename=self.env.conf_path,
#                     use_sudo=True)
            else:
#                 r.sudo(
#                     "sed '/script.engine.groovy.inline.search: on/d' {conf_path}"\
#                         .format(conf_path=self.env.conf_path))
                r.sed(
                    filename=self.env.conf_path,
                    before='script.engine.groovy.inline.search: on',
                    after='script.engine.groovy.inline.search: off',
                    use_sudo=True,
                )

            self.enable()
            self.restart()
        else:
            self.disable()
            self.stop()

elasticsearch = ElasticSearchSatchel()



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class MemcachedSatchel(ServiceSatchel):

    name = 'memcached'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['memcached'],
        }

    def set_defaults(self):
        super().set_defaults()
        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start memcached',
                UBUNTU: 'service memcached start',
            },
            STOP:{
                FEDORA: 'systemctl stop memcached',
                UBUNTU: 'service memcached stop',
            },
            STATUS:{
                FEDORA: 'systemctl status memcached',
                UBUNTU: 'service memcached status',
            },
            DISABLE:{
                FEDORA: 'systemctl disable memcached',
                UBUNTU: 'chkconfig memcached off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable memcached',
                UBUNTU: 'chkconfig memcached on',
            },
            RELOAD:{
                FEDORA: 'systemctl reload memcached',
                UBUNTU: 'service memcached reload',
            },
            RESTART:{
                FEDORA: 'systemctl restart memcached',
                #UBUNTU: 'service memcached restart',
                # Note, the sleep 5 is necessary because the stop/start appears to
                # happen in the background but gets aborted if Fabric exits before
                # it completes.
                UBUNTU: 'service memcached restart; sleep 3',
            },
        }

    @task(precursors=['packager'])
    def configure(self):
        pass



"""
Celery component.

Note, we manage Celery through Supervisor, since it's extremely difficult to
run several instances of Celery for multiple Apache sites.

"""
from pprint import pprint

from burlap.constants import *
from burlap import ServiceSatchel
from burlap.decorators import task

class CelerySatchel(ServiceSatchel):

    name = 'celery'

    @property
    def packager_system_packages(self):
        d = {}
        if self.env.use_system_packages:
            d.update({
                FEDORA: ['celery', 'django-celery'],
                UBUNTU: ['celery', 'django-celery'],
            })
        return d

    def set_defaults(self):

        # Better versions are available in PyPI, so don't use system packages by default.
        self.env.use_system_packages = False

        self.env.config_path = '/etc/sysconfig/celeryd'
        self.env.daemon_opts = '--concurrency=1 --loglevel=DEBUG'
        self.env.daemon_path = '/etc/init.d/celeryd'
        self.env.log_path_template = '/var/log/celeryd-{SITE}.log'
        self.env.celerybeat_log_path_template = '/var/log/celerybeat-{SITE}.log'
        self.env.celeryd_command = 'celeryd'
        self.env.has_worker = False
        self.env.daemon_user = 'www-data'
        self.env.numprocs = 1
        self.env.force_stop_command = 'pkill -9 -f celery'
        self.env.celeryd_command_template = None
        self.env.supervisor_directory_template = None#'/usr/local/myproject'

        #DEPRECATED
        self.env.has_celerybeat = False
        self.env.celerybeat_command = 'celerybeat'
        self.env.paths_owned = ['/tmp/celerybeat-schedule*', '/var/log/celery*']
        self.env.celerybeat_opts_template = ('--schedule=/tmp/celerybeat-schedule-{SITE} --pidfile=/tmp/celerybeat-{SITE}.pid '
            '--logfile={celery_celerybeat_log_path} --loglevel=DEBUG')
        self.env.celerybeat_command_template = ('{celery_supervisor_python} {celery_supervisor_django_manage} '
            '{celery_celerybeat_command} {celery_celerybeat_opts}')

        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start celeryd.service',
                UBUNTU: 'service celeryd start',
            },
            STOP:{
                FEDORA: 'systemctl stop celery.service',
                UBUNTU: 'service celeryd stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable httpd.service',
                UBUNTU: 'chkconfig celeryd off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable httpd.service',
                UBUNTU: 'chkconfig celeryd on',
            },
            RESTART:{
                FEDORA: 'systemctl stop celeryd.service; pkill -9 -f celery; systemctl start celeryd.service',
                UBUNTU: 'service celeryd stop; pkill -9 -f celery; service celeryd start',
            },
            STATUS:{
                FEDORA: 'systemctl status celeryd.service',
                UBUNTU: 'service celeryd status',
            },
        }

    @task
    def purge(self):
        """
        Clears all pending tasks in the Celery queue.
        """
        self.render_paths()
        r = self.local_renderer
        r.sudo('export SITE={SITE}; export ROLE={ROLE}; {celery_supervisor_django_manage} celeryctl purge')

    @task
    def force_stop(self):
        """
        Forcibly terminates all Celery processes.
        """
        r = self.local_renderer
        with self.settings(warn_only=True):
            r.sudo('pkill -9 -f celery')
        r.sudo('rm -f /tmp/celery*.pid')

    @task
    def set_permissions(self):
        """
        Sets ownership and permissions for Celery-related files.
        """
        r = self.local_renderer
        for path in r.env.paths_owned:
            r.env.path_owned = path
            r.sudo('chown {celery_daemon_user}:{celery_daemon_user} {celery_path_owned}')

    @task
    def render_paths(self):
        r = self.local_renderer
        r.env.supervisor_directory = r.format(r.env.supervisor_directory_template)
        r.env.celeryd_command = r.format(r.env.celeryd_command_template or r.env.celeryd_command)
        r.env.log_path = r.format(r.env.log_path_template)

    @task
    def create_supervisor_services(self, site):
        """
        This is called for each site to render a Celery config file.
        """

        self.vprint('create_supervisor_services:', site)

        # Since this may be called for multiple sites, ensure we don't pollute our renderer with values from
        # each site, and instead get a fresh renderer.
        self.clear_local_renderer()

        self.set_site_specifics(site=site)

        r = self.local_renderer
        if self.verbose:
            print('r.env:')
            pprint(r.env, indent=4)

        self.vprint('r.env.has_worker:', r.env.has_worker)
        if not r.env.has_worker:
            self.vprint('skipping: no celery worker')
            return

        if self.name.lower() not in self.genv.services:
            self.vprint('skipping: celery not enabled')
            return

        hostname = self.current_hostname
        target_sites = self.genv.available_sites_by_host.get(hostname, None)
        if target_sites and site not in target_sites:
            self.vprint('skipping: site not supported on this server')
            return

        self.render_paths()

        conf_name = 'celery_%s.conf' % site
        ret = r.render_to_string('celery/celery_supervisor.template.conf')
        return conf_name, ret

    @task(post_callback=True)
    def register_callbacks(self):
        from burlap.supervisor import supervisor
        supervisor.register_callback(self.create_supervisor_services)

    @task(precursors=['packager'])
    def configure(self):
        pass

celery = CelerySatchel()



from fabric.api import settings

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class NetworkManagerSatchel(ServiceSatchel):
    """
    Configures Network Manager for automatically controlling network interfaces.

    https://fedoraproject.org/wiki/Networking/CLI#Wifi
    """

    name = 'nm'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['network-manager', 'cron'],
            DEBIAN: ['network-manager', 'cron'],
        }

    templates = (
        'check_networkmanager.sh',
        'etc_crond_check_networkmanager',
    )

    def set_defaults(self):
        self.env.check_enabled = False

        self.env.service_commands = {
            START:{
                UBUNTU: 'service network-manager start',
                DEBIAN: 'service network-manager start',
            },
            STOP:{
                UBUNTU: 'service network-manager stop',
                DEBIAN: 'service network-manager stop',
            },
            DISABLE:{
                DEBIAN: 'update-rc.d network-manager disable',
                #UBUNTU: 'chkconfig network-manager off',
                UBUNTU: 'systemctl disable network-manager.service',
                (UBUNTU, '14.04'): 'echo "manual" | sudo tee /etc/init/network-manager.override',
            },
            ENABLE:{
                DEBIAN: 'update-rc.d network-manager enable',
                #UBUNTU: 'chkconfig network-manager on',
                UBUNTU: 'systemctl enable network-manager.service',
                (UBUNTU, '14.04'): 'rm /etc/init/network-manager.override || true',
            },
            RESTART:{
                UBUNTU: 'service network-manager restart',
                DEBIAN: 'service network-manager restart',
            },
            STATUS:{
                UBUNTU: 'service network-manager status',
                DEBIAN: 'service network-manager status',
            },
        }

        self.env.check_script_path = '/usr/local/bin/check_networkmanager.sh'
        self.env.cron_script_path = '/etc/cron.d/check_networkmanager'
        self.env.check_script_template = '%s/check_networkmanager.sh' % self.name
        self.env.cron_perms = '600'

        self.env.connections = {} # {ssid: passphrase}

    @task
    def add_wifi_connection(self, ssid, passphrase=None):
        r = self.local_renderer
        r.env.ssid = ssid = ssid
        r.env.passphrase = passphrase
        r.sudo('nmcli device wifi connect "{ssid}" password "{passphrase}"')

    @task
    def remove_connection(self, ssid):
        r = self.local_renderer
        r.env.ssid = ssid = ssid
        #r.sudo("nmcli connection delete `nmcli --fields NAME,UUID con list | grep -i {ssid} | awk '{print $2}'`")
        r.sudo("nmcli connection delete id {ssid}")

    @task
    def dev_status(self):
        r = self.local_renderer
        r.sudo('nmcli device status')

    @task
    def dev_wifi_list(self):
        r = self.local_renderer
        r.sudo('nmcli device wifi list')

    @task
    def configure_checker(self):
        r = self.local_renderer
        if r.env.check_enabled:
            # Installs a crontab to check Network-Manager every ten minutes
            # and restart it if theres' no Internet connection.
            self.install_script(
                local_path=r.env.check_script_template,
                remote_path=self.lenv.check_script_path)
            remote_path = r.put(
                local_path=self.find_template('%s/etc_crond_check_networkmanager' % self.name),
                remote_path=self.env.cron_script_path, use_sudo=True)[0]
            r.sudo('chown root:root %s' % remote_path)#env.put_remote_path)
            # Must be 600, otherwise gives INSECURE MODE error.
            # http://unix.stackexchange.com/questions/91202/cron-does-not-print-to-syslog
            r.sudo('chmod %s %s' % (self.env.cron_perms, remote_path))#env.put_remote_path)
            r.sudo('service cron restart')
        else:
            r.sudo('rm -f {cron_script_path}'.format(**self.lenv))
            r.sudo('service cron restart')

    @task(precursors=['packager', 'user', 'cron'])
    def configure(self):

        r = self.local_renderer

        lm = self.last_manifest
        lm_connections = lm.connections or {}

        if r.env.enabled:

            # Clear the /etc/network/interfaces so NM will control all interfaces.
            if not self.files.exists('/etc/network/interfaces'):
                r.sudo('mv /etc/network/interfaces /etc/network/interfaces.bak')
            r.sudo('rm -f /etc/network/interfaces')
            r.sudo('touch /etc/network/interfaces')
            r.sudo('echo -e "auto lo\\niface lo inet loopback" > /etc/network/interfaces')

            self.enable()
            self.restart()
        else:
            self.disable()
            self.stop()

        self.configure_checker()

        # When enabling wifi for the first time, a reboot is required.
        if not lm_connections and r.env.connections and r.env.enabled:
            r.reboot(wait=300, timeout=60)

        # Remove deleted connections.
        # Note, this will fail if the host's connection is currently using the connection.
        for old_ssid in lm_connections:
            if old_ssid not in r.env.connections:
                self.remove_connection(old_ssid)

        # Add or update connections.
        for ssid, passphrase in r.env.connections.items():

            # Remove old connection.
            if ssid in lm_connections and lm_connections[ssid] != passphrase:
                self.remove_connection(ssid)

            # Add new connection.
            with settings(warn_only=True):
                # This may through an error code, but on reboot, the connection will appear and be connected.
                self.add_wifi_connection(ssid, passphrase)

nm = NetworkManagerSatchel()



import os
import re
from datetime import datetime, date

import dateutil.parser
import pytz

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task
#from burlap import common

# import dateutil.parser
# import pytz
from fabric.api import runs_once, hide

class SSLSatchel(ServiceSatchel):

    name = 'sslhelper'

    def set_defaults(self):
        self.env.country = '?'
        self.env.state = '?'
        self.env.city = '?'
        self.env.organization = '?'
        self.env.common_name = '?'
        self.env.days = 365
        self.env.length = 4096
        self.env.domain = ''
        self.env.self_signed_certificate_command = 'openssl req -new -newkey rsa:{sslhelper_length} ' \
            '-days {sslhelper_days} -nodes -x509 ' \
            '-subj "/C={sslhelper_country}/ST={sslhelper_state}/L={sslhelper_city}/O={sslhelper_organization}/' \
            'CN={domain}" ' \
            '-keyout {sslhelper_base_dst}.key -out {sslhelper_base_dst}.crt'
        self.env.generate_csr_command = 'openssl req -nodes -newkey rsa:{sslhelper_length} ' \
            '-subj "/C={sslhelper_country}/ST={sslhelper_state}/L={sslhelper_city}/O={sslhelper_organization}/' \
            'CN={domain}" ' \
            '-keyout {sslhelper_base_dst}.{sslhelper_csr_year}.key -out {sslhelper_base_dst}.{sslhelper_csr_year}.csr'

    @task
    def generate_self_signed_certificate(self, domain='', r=None):
        """
        Generates a self-signed certificate for use in an internal development
        environment for testing SSL pages.

        http://almostalldigital.wordpress.com/2013/03/07/self-signed-ssl-certificate-for-ec2-load-balancer/
        """
        r = self.local_renderer
        r.env.domain = domain or r.env.domain
        assert r.env.domain, 'No SSL domain defined.'
        role = r or self.genv.ROLE or ALL
        ssl_dst = 'roles/%s/ssl' % (role,)
        if not os.path.isdir(ssl_dst):
            os.makedirs(ssl_dst)
        r.env.base_dst = '%s/%s' % (ssl_dst, r.env.domain)
        r.local(r.env.self_signed_certificate_command)

    @task
    @runs_once
    def generate_csr(self, domain='', r=None):
        """
        Creates a certificate signing request to be submitted to a formal
        certificate authority to generate a certificate.

        Note, the provider may say the CSR must be created on the target server,
        but this is not necessary.
        """
        r = r or self.local_renderer
        r.env.domain = domain or r.env.domain
        role = self.genv.ROLE or ALL
        site = self.genv.SITE or self.genv.default_site
        ssl_dst = 'roles/%s/ssl' % (role,)
        if not os.path.isdir(ssl_dst):
            os.makedirs(ssl_dst)
        for site, site_data in self.iter_sites():
            assert r.env.domain, 'No SSL domain defined.'
            r.env.sslhelper_base_dst = '%s/%s' % (ssl_dst, r.env.domain.replace('*.', ''))
            r.env.sslhelper_csr_year = date.today().year
            r.local(r.env.generate_csr_command)

    def get_expiration_date(self, fn):
        """
        Reads the expiration date of a local crt file.
        """
        r = self.local_renderer
        r.env.crt_fn = fn
        with hide('running'):
            ret = r.local('openssl x509 -noout -in {sslhelper_crt_fn} -dates', capture=True)
        matches = re.findall('notAfter=(.*?)$', ret, flags=re.IGNORECASE)
        if matches:
            return dateutil.parser.parse(matches[0])

    @task
    def list_expiration_dates(self, base='roles/all/ssl'):
        """
        Scans through all local .crt files and displays the expiration dates.
        """
        max_fn_len = 0
        max_date_len = 0
        data = []
        for fn in os.listdir(base):
            fqfn = os.path.join(base, fn)
            if not os.path.isfile(fqfn):
                continue
            if not fn.endswith('.crt'):
                continue
            expiration_date = self.get_expiration_date(fqfn)
            max_fn_len = max(max_fn_len, len(fn))
            max_date_len = max(max_date_len, len(str(expiration_date)))
            data.append((fn, expiration_date))
        print('%s %s %s' % ('Filename'.ljust(max_fn_len), 'Expiration Date'.ljust(max_date_len), 'Expired'))
        now = datetime.now().replace(tzinfo=pytz.UTC)
        for fn, dt in sorted(data):
            if dt is None:
                expired = '?'
            elif dt < now:
                expired = 'YES'
            else:
                expired = 'NO'
            print('%s %s %s' % (fn.ljust(max_fn_len), str(dt).ljust(max_date_len), expired))

    @task
    def verify_certificate_chain(self, base=None, crt=None, csr=None, key=None):
        """
        Confirms the key, CSR, and certificate files all match.
        """
        from burlap.common import get_verbose, print_fail, print_success

        r = self.local_renderer

        if base:
            crt = base + '.crt'
            csr = base + '.csr'
            key = base + '.key'
        else:
            assert crt and csr and key, 'If base not provided, crt and csr and key must be given.'

        assert os.path.isfile(crt)
        assert os.path.isfile(csr)
        assert os.path.isfile(key)

        csr_md5 = r.local('openssl req -noout -modulus -in %s | openssl md5' % csr, capture=True)
        key_md5 = r.local('openssl rsa -noout -modulus -in %s | openssl md5' % key, capture=True)
        crt_md5 = r.local('openssl x509 -noout -modulus -in %s | openssl md5' % crt, capture=True)

        match = crt_md5 == csr_md5 == key_md5

        if self.verbose or not match:
            print('crt:', crt_md5)
            print('csr:', csr_md5)
            print('key:', key_md5)

        if match:
            print_success('Files look good!')
        else:
            print_fail('Files no not match!')
            raise Exception('Files no not match!')

    @task
    def configure(self):
        pass

sslhelper = SSLSatchel()



"""
NTP component.

Merely a stub to document which packages should be installed
if a system uses this component.

It should be otherwise maintenance-free and have required settings to configure.
"""
from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class NTPClientSatchel(ServiceSatchel):

    name = 'ntpclient'

    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['ntpdate', 'ntp'],
            UBUNTU: ['ntpdate', 'ntp'],
            (UBUNTU, '12.04'): ['ntpdate', 'ntp'],
            (UBUNTU, '14.04'): ['ntpdate', 'ntp'],
            (UBUNTU, '16.04'): ['ntpdate', 'ntp'],
            DEBIAN: ['ntpdate', 'ntp'],
        }

    def set_defaults(self):
        self.env.enabled = True
        self.env.default_server = 'time.nist.gov'
        self.env.service_commands = {
            START:{
                UBUNTU: 'service ntp start',
                DEBIAN: 'service ntp start',
            },
            STOP:{
                UBUNTU: 'service ntp stop',
                DEBIAN: 'service ntp stop',
            },
            DISABLE:{
                UBUNTU: 'update-rc.d -f ntp remove',
                DEBIAN: 'update-rc.d ntp disable',
            },
            ENABLE:{
                UBUNTU: 'update-rc.d ntp defaults',
                DEBIAN: 'update-rc.d ntp enable',
            },
            RESTART:{
                UBUNTU: 'service ntp restart',
                DEBIAN: 'service ntp restart',
            },
            STATUS:{
                UBUNTU: 'service ntp status',
                DEBIAN: 'service ntp status',
            },
        }

    @task
    def force_update(self):
        r = self.local_renderer
        r.sudo('service ntp stop')
        r.sudo('ntpdate {default_server}')
        r.sudo('service ntp start')

    @task(precursors=['packager', 'user'])
    def configure(self):
        if self.env.enabled:
            self.install_packages()
            self.enable()
            self.restart()
        else:
            self.disable()
            self.stop()

ntpclient = NTPClientSatchel()



import sys
import re
import traceback
from datetime import date
from pprint import pprint

from burlap import ContainerSatchel
from burlap.constants import *
from burlap.decorators import task


class JiraHelperSatchel(ContainerSatchel):

    name = 'jirahelper'
    _jira = None

    def set_defaults(self):

        self.env.server = None
        self.env.basic_auth_username = None
        self.env.basic_auth_password = None
        self.env.update_from_git = False
        self.env.ticket_update_message_template = 'This has been deployed to {role}.'

        # A map of status->transition to follow when making deployments.
        self.env.deploy_workflow = {}

        # A map of the new-status->new-assignee to auto-assign.
        self.env.assignee_by_status = {}

        # The regex used to search the commit messages for a ticket number.
        self.env.ticket_pattern = None

        # If True, update release (fixVersion field) for all deployed issues in update_tickets_from_git().
        self.env.update_ticket_release = False
        self.env.release_format = '%Y.%m.%d'

    @property
    def project(self):
        """
        'PROJECT-1234' -> 'PROJECT'
        """
        return self.env.ticket_pattern.split('-', maxsplit=1)[0]

    @property
    def jira(self):
        """
        Attach a JIRA instance to JiraHelperSatchel class to prevent unnecessary re-initialization.
        """
        from jira import JIRA

        if self._jira is None:
            self.vprint('Connecting to %s with user %s...' % (self.env.server, self.env.basic_auth_username))
            self._jira = JIRA({
                'server': self.env.server
            }, basic_auth=(self.env.basic_auth_username, self.env.basic_auth_password))
        return self._jira

    def get_tickets_from_str(self, s):
        """
        From a string, get all Jira issue keys matching a pattern.

        Args:
            s (str): String containing zero or more Jira issue keys

        Returns:
            tickets (set): Jira issue key strings
        """
        if self.env.ticket_pattern:
            pattern = re.compile(self.env.ticket_pattern, flags=re.IGNORECASE)
            tickets = pattern.findall(s)
        else:
            tickets = []
        return {ticket.strip().upper() for ticket in tickets}

    @task
    def get_tickets_between_commits(self, old_commit, new_commit):
        """
        Retrieve all Jira issue linked to commits between the given commit numbers on the current branch.

        Args:
            old_commit (str): Hash of ancestor git commit
            new_commit (str): Hash of descendant git commit

        Returns:
            tickets (set): Jira issue key strings corresponding to the new commits
        """
        from burlap.git import gittracker

        if self.env.ticket_pattern:
            commit_logs = gittracker.get_logs_between_commits(old_commit, new_commit)
            tickets = self.get_tickets_from_str(commit_logs)
        else:
            tickets = set()
        self.vprint(tickets)
        return tickets

    @task
    def test_connection(self):
        from burlap.common import print_success, print_fail

        try:
            result = self.jira.search_issues('status=resolved')
            self.vprint('result:', result)
            print_success('OK')
        except Exception as exc:
            print_fail('ERROR: %s' % exc)

    def get_release(self):
        """
        Get or create current Jira release for project, using env.release_format.

        Returns:
            release (str): Name of release.
        """
        release = date.today().strftime(self.env.release_format)
        self.vprint('Looking for release %s...' % release)
        if not self.jira.get_project_version_by_name(self.project, release):
            self.vprint('Creating release %s...' % release)
            self.jira.create_version(release, project=self.project, releaseDate=date.today().strftime('%Y-%m-%d'))

        return release

    @task
    def update_ticket_from_git(self, ticket, release=None):
        """
        Update ticket status and release in Jira.

        Args:
            ticket (str): Jira issue key
            release (str or None): Name of release to associate with issue
        """
        from jira import JIRA, JIRAError

        r = self.local_renderer

        # Mention this Jira update.
        r.env.role = r.genv.ROLE.lower()
        comment = r.format(self.env.ticket_update_message_template)
        print('Commenting on ticket %s: %s' % (ticket, comment))
        if not self.dryrun:
            self.jira.add_comment(ticket, comment)

        print('Looking up Jira ticket %s...' % ticket)
        issue = self.jira.issue(ticket)
        self.vprint('Ticket %s retrieved.' % ticket)

        # Update ticket release.
        if self.env.update_ticket_release:
            try:
                if not release:
                    release = self.get_release()
                self.vprint('Updating issue %s to release %s.' % (issue, release))
                issue.update(fields={'fixVersions': [{'name': release}]})
            except JIRAError as exc:
                print(exc)

        # Update ticket status.
        recheck = False
        for i in range(10):
            transition_to_id = {t['name']: t['id'] for t in self.jira.transitions(issue)}
            self.vprint('%i allowable transitions found:' % len(transition_to_id))
            if self.verbose:
                pprint(transition_to_id)
            self.vprint('issue.fields.status.id:', issue.fields.status.id)
            self.vprint('issue.fields.status.name:', issue.fields.status.name)
            jira_status_name = issue.fields.status.name
            self.vprint('jira_status_name:', jira_status_name)
            next_transition_name = self.env.deploy_workflow.get(jira_status_name)
            self.vprint('next_transition_name:', next_transition_name)
            next_transition_id = transition_to_id.get(next_transition_name)
            self.vprint('next_transition_id:', next_transition_id)
            if next_transition_name:
                # Note: assignment should happen after transition, since the assignment may
                # remove transitions that we need.
                print('Updating ticket %s to status %s (%s).' % (ticket, next_transition_name, next_transition_id))
                if next_transition_id and not self.dryrun:
                    try:
                        self.jira.transition_issue(issue, next_transition_id)
                        recheck = True
                    except AttributeError as e:
                        print('Unable to transition ticket %s to %s: %s' % (ticket, next_transition_name, e),
                              file=sys.stderr)
                        traceback.print_exc()

                # Get new assignee by status.
                new_assignee = self.env.assignee_by_status.get(next_transition_name)
                if new_assignee == 'reporter':
                    new_assignee = issue.fields.reporter

                if new_assignee:
                    print('Assigning ticket %s to %s.' % (ticket, new_assignee.displayName))
                    if not self.dryrun:
                        try:
                            self.jira.assign_issue(issue, new_assignee)
                        except JIRAError as e:
                            print('Unable to reassign ticket %s to %s: %s' % (ticket, new_assignee.displayName, e),
                                  file=sys.stderr)
                else:
                    print('No new assignee found.')
            else:
                recheck = False
                print(
                    'No transitions found for ticket %s currently in status "%s".' % (ticket, issue.fields.status.name)
                )

            if not recheck:
                break

    @task
    def update_tickets_from_git(self, from_commit=None, to_commit=None):
        """
        Find all ticket numbers and update their status in Jira.

        Run during a deployment.
        Looks at all commits between now and the last deployment.
        """
        from burlap.git import gittracker, CURRENT_COMMIT

#         get_current_commit = gittracker.get_current_commit
#         GITTRACKER = gittracker.name.upper()

        # Ensure this is only run once per role.
        if self.genv.host_string != self.genv.hosts[-1]:
            self.vprint('Not first server. Aborting.')
            return

        self.vprint('self.env.update_from_git:', self.env.update_from_git)
        self.vprint('self.genv.jirahelper_update_from_git:', self.genv.jirahelper_update_from_git)
        if not self.env.update_from_git:
            self.vprint('Update from git disabled. Aborting.')
            return

        if not self.env.ticket_pattern:
            self.vprint('No ticket pattern defined. Aborting.')
            return

        if not self.env.basic_auth_username or not self.env.basic_auth_password:
            self.vprint('Username or password not given. Aborting.')
            return

        # During a deployment, we should be given these, but for testing, look up the diffs dynamically.
        last = gittracker.last_manifest
        current = gittracker.current_manifest

        last_commit = from_commit or last.current_commit#[CURRENT_COMMIT]
        print('last_commit:', last_commit)
        current_commit = to_commit or current[CURRENT_COMMIT]
        print('current_commit:', current_commit)

        if not last_commit or not current_commit:
            print('Missing commit ID. Aborting.')
            return

        self.vprint('-'*80)
        self.vprint('last.keys:', last.keys())
        self.vprint('-'*80)
        self.vprint('current.keys:', current.keys())

#         try:
#             last_commit = last['GITTRACKER']['current_commit']
#         except KeyError:
#             return
#         current_commit = current['GITTRACKER']['current_commit']

        # Find all tickets deployed between last deployment and now.
        tickets = self.get_tickets_between_commits(last_commit, current_commit)
        self.vprint('tickets:', tickets)

        if not tickets:
            return

        # Get or create Jira release for tickets.
        if self.env.update_ticket_release:
            release = self.get_release()
        else:
            release = None

        # Update all tickets in Jira.
        for ticket in tickets:
            try:
                self.update_ticket_from_git(ticket, release=release)
            except Exception:
                traceback.print_exc()

        # Mark Jira release as released.
        if self.env.update_ticket_release:
            self.vprint('Releasing version %s...:', release)
            release_obj = self.jira.get_project_version_by_name(self.project, release)
            release_obj.update(released=True)

jirahelper = JiraHelperSatchel()



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class InadynSatchel(ServiceSatchel):

    name = 'inadyn'

    def set_defaults(self):

        self.env.daemon_name = 'inadyn'
        self.env.config_user = 'root'
        self.env.config_group = 'debian-inadyn'
        self.env.config_chmod = 640
        self.env.config_path = '/etc/inadyn.conf'
        self.env.system = None
        self.env.dyndns_system = None
        self.env.update_period_sec = 600
        self.env.username = None
        self.env.password = None
        self.env.aliases_by_hostname = {}
        self.env.check_command_template = 'inadyn ' \
            '--username {username} --password "{password}" ' \
            '--update_period_sec {update_period_sec} --dyndns_system {dyndns_system} --alias {alias} --test'

        self.env.service_commands = {
            START:{
                UBUNTU: 'service %s start' % self.env.daemon_name,
            },
            STOP:{
                UBUNTU: 'service %s stop' % self.env.daemon_name,
            },
            DISABLE:{
                UBUNTU: 'systemctl disable %s.service' % self.env.daemon_name,
            },
            ENABLE:{
                UBUNTU: 'systemctl enable %s.service' % self.env.daemon_name,
            },
            RESTART:{
                UBUNTU: 'service %s restart' % self.env.daemon_name,
            },
            STATUS:{
                UBUNTU: 'service %s status' % self.env.daemon_name,
            },
        }

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['inadyn'],
        }

    def _validate_settings(self):
        print('_validate_settings.1')
        r = self.local_renderer
        assert r.env.username, 'No username.'
        print('_validate_settings.1:', r.env.username)
        assert r.env.password, 'No password.'
        assert r.env.dyndns_system, 'No dyndns_system.'
        hn = self.current_hostname
        r.env.aliases = r.env.aliases_by_hostname.get(hn)
        print('_validate_settings.2:', r.env.aliases)
        assert r.env.aliases, 'No aliases for hostname %s.' % hn
        assert isinstance(r.env.aliases, (list, tuple)), 'Aliases must be list.'
        return True

    @task
    def check(self):
        """
        Run inadyn from the commandline to test the configuration.

        To be run like:

            fab role inadyn.check

        """
        self._validate_settings()
        r = self.local_renderer
        r.env.alias = r.env.aliases[0]
        r.sudo(r.env.check_command_template)

    @task
    def write_config(self):
        self._validate_settings()
        r = self.local_renderer
        r.env.remote_path = r.env.config_path
        self.genv.inadyn_aliases = r.env.aliases
        r.install_config(local_path='inadyn/etc_inadyn_conf.template', remote_path=r.env.config_path)
        r.sudo('chown {config_user}:{config_group} {config_path}')
        r.sudo('chmod {config_chmod} {config_path}')
        del self.genv['inadyn_aliases']
        r.install_config(local_path='inadyn/etc_default_inadyn.template', remote_path='/etc/default/inadyn')

    @task(precursors=['packager', 'user'])
    def configure(self):
        if self.env.enabled:
            self.enable()
            self.write_config()
            self.restart()
        else:
            self.disable()
            self.stop()

inadyn = InadynSatchel()



"""
PostgreSQL users and databases
==============================

This module provides tools for creating PostgreSQL users and databases.

"""
import os

import six

from fabric.api import cd, hide, sudo, settings

from burlap import Satchel
from burlap.constants import *
from burlap.db import DatabaseSatchel
from burlap.decorators import task

POSTGIS = 'postgis'
POSTGRESQL = 'postgresql'

def _run_as_pg(command):
    """
    Run command as 'postgres' user
    """
    with cd('~postgres'):
        return sudo(command, user='postgres')


def user_exists(name):
    """
    Check if a PostgreSQL user exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'),
                  warn_only=True):
        res = _run_as_pg('''psql -t -A -c "SELECT COUNT(*) FROM pg_user WHERE usename = '%(name)s';"''' % locals())
    return (res == "1")


def create_user(name, password, superuser=False, createdb=False,
                createrole=False, inherit=True, login=True,
                connection_limit=None, encrypted_password=False):
    """
    Create a PostgreSQL user.

    Example::

        import burlap

        # Create DB user if it does not exist
        if not burlap.postgres.user_exists('dbuser'):
            burlap.postgres.create_user('dbuser', password='somerandomstring')

        # Create DB user with custom options
        burlap.postgres.create_user('dbuser2', password='s3cr3t',
            createdb=True, createrole=True, connection_limit=20)

    """
    options = [
        'SUPERUSER' if superuser else 'NOSUPERUSER',
        'CREATEDB' if createdb else 'NOCREATEDB',
        'CREATEROLE' if createrole else 'NOCREATEROLE',
        'INHERIT' if inherit else 'NOINHERIT',
        'LOGIN' if login else 'NOLOGIN',
    ]
    if connection_limit is not None:
        options.append('CONNECTION LIMIT %d' % connection_limit)
    password_type = 'ENCRYPTED' if encrypted_password else 'UNENCRYPTED'
    options.append("%s PASSWORD '%s'" % (password_type, password))
    options = ' '.join(options)
    _run_as_pg('''psql -c "CREATE USER %(name)s %(options)s;"''' % locals())


def drop_user(name):
    """
    Drop a PostgreSQL user.

    Example::

        import burlap

        # Remove DB user if it exists
        if burlap.postgres.user_exists('dbuser'):
            burlap.postgres.drop_user('dbuser')

    """
    _run_as_pg('''psql -c "DROP USER %(name)s;"''' % locals())


def database_exists(name):
    """
    Check if a PostgreSQL database exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'),
                  warn_only=True):
        return _run_as_pg('''psql -d %(name)s -c ""''' % locals()).succeeded


def create_database(name, owner, template='template0', encoding='UTF8',
                    locale='en_US.UTF-8'):
    """
    Create a PostgreSQL database.

    Example::

        import burlap

        # Create DB if it does not exist
        if not burlap.postgres.database_exists('myapp'):
            burlap.postgres.create_database('myapp', owner='dbuser')

    """
    _run_as_pg('''createdb --owner %(owner)s --template %(template)s \
                  --encoding=%(encoding)s --lc-ctype=%(locale)s \
                  --lc-collate=%(locale)s %(name)s''' % locals())


def create_schema(name, database, owner=None):
    """
    Create a schema within a database.
    """
    if owner:
        _run_as_pg('''psql %(database)s -c "CREATE SCHEMA %(name)s AUTHORIZATION %(owner)s"''' % locals())
    else:
        _run_as_pg('''psql %(database)s -c "CREATE SCHEMA %(name)s"''' % locals())


class PostgreSQLSatchel(DatabaseSatchel):
    """
    Represents a PostgreSQL server.
    """

    name = 'postgresql'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['postgresql-11'],
            (UBUNTU, '12.04'): ['postgresql-9.1'],
            (UBUNTU, '14.04'): ['postgresql-9.3'],
            (UBUNTU, '16.04'): ['postgresql-10'],
            (UBUNTU, '18.04'): ['postgresql-11'],
            (UBUNTU, '20.04'): ['postgresql-13'],
        }

    def set_defaults(self):
        super().set_defaults()

        # Note, if you use gzip, you can't use parallel restore.
        #self.env.dump_command = 'time pg_dump -c -U {db_user} --no-password --blobs --format=c --schema=public --host={db_host} {db_name} > {dump_fn}'
        self.env.dump_command = 'time pg_dump -c -U {db_user} --no-password --blobs --format=c --schema=public --host={db_host} --dbname={db_name} | ' \
            'gzip -c > {dump_fn}'
        self.env.dump_fn_template = '{dump_dest_dir}/db_{db_type}_{SITE}_{ROLE}_{db_name}_$(date +%Y%m%d).sql.gz'

        #self.env.load_command = 'gunzip < {remote_dump_fn} | pg_restore --jobs=8 -U {db_root_username} --format=c --create --dbname={db_name}'
        self.env.load_command = 'gunzip < {remote_dump_fn} | ' \
            'pg_restore -U {db_root_username} --host={db_host} --format=c --clean --if-exists --schema=public --dbname={db_name}'

        self.env.createlangs = ['plpgsql'] # plpythonu
        self.env.postgres_user = 'postgres'
        self.env.encoding = 'UTF8'
        self.env.locale = 'en_US.UTF-8'
        self.env.custom_load_cmd = ''
        self.env.port = 5432
        self.env.pgpass_path = '~/.pgpass'
        self.env.pgpass_chmod = 600
        self.env.force_version = None
        self.env.version_command = r'`psql --version | grep -m 1 -o -E "[0-9]+\.[0-9]+" | head -1`'
        self.env.engine = POSTGRESQL # 'postgresql' | postgis

        # If true, treats the current database as existing as a schema in a multi-tenant environment,
        # so commands to drop/create the entire database will be replaced with commands to drop/create a schema.
        self.env.schema_mt = False

        self.env.db_root_username = 'postgres'
        self.env.db_root_password = 'password'

        self.env.apt_repo_enabled = False

        # https://askubuntu.com/questions/831292/how-to-install-postgresql-9-6-on-any-ubuntu-version
        # Populated from https://www.postgresql.org/download/linux/ubuntu/
        self.env.apt_repo = None

        self.env.apt_key = 'https://www.postgresql.org/media/keys/ACCC4CF8.asc'

        self.env.service_commands = {
            START:{
                UBUNTU: 'service postgresql start',
            },
            STOP:{
                UBUNTU: 'service postgresql stop',
            },
            ENABLE:{
                UBUNTU: 'update-rc.d postgresql defaults',
            },
            DISABLE:{
                UBUNTU: 'update-rc.d -f postgresql remove',
            },
            RESTART:{
                UBUNTU: 'service postgresql restart',
            },
            STATUS:{
                UBUNTU: 'service postgresql status',
            },
        }

    #https://askubuntu.com/questions/831292/how-to-install-postgresql-9-6-on-any-ubuntu-version
    @property
    def packager_repositories(self):
        ver = self.os_version
        if ver.type == LINUX:
            if ver.distro == UBUNTU:
                if ver.release == '16.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ xenial-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
                if ver.release == '18.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ bionic-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
                if ver.release == '20.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ focal-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
        raise NotImplementedError

    @task
    def write_pgpass(self, name=None, site=None, use_sudo=0, root=0):
        """
        Write the file used to store login credentials for PostgreSQL.
        """

        r = self.database_renderer(name=name, site=site)

        root = int(root)

        use_sudo = int(use_sudo)

        r.run('touch {pgpass_path}')
        if '~' in r.env.pgpass_path:
            r.run('chmod {pgpass_chmod} {pgpass_path}')
        else:
            r.sudo('chmod {pgpass_chmod} {pgpass_path}')

        if root:
            r.env.shell_username = r.env.get('db_root_username', 'postgres')
            r.env.shell_password = r.env.get('db_root_password', 'password')
        else:
            r.env.shell_username = r.env.db_user
            r.env.shell_password = r.env.db_password

        r.append(
            '{db_host}:{port}:*:{shell_username}:{shell_password}',
            r.env.pgpass_path,
            use_sudo=use_sudo)

    @task
    def dumpload(self, site=None, role=None):
        """
        Dumps and loads a database snapshot simultaneously.
        Requires that the destination server has direct database access
        to the source server.

        This is better than a serial dump+load when:
        1. The network connection is reliable.
        2. You don't need to save the dump file.

        The benefits of this over a dump+load are:
        1. Usually runs faster, since the load and dump happen in parallel.
        2. Usually takes up less disk space since no separate dump file is
            downloaded.
        """
        r = self.database_renderer(site=site, role=role)
        r.run('pg_dump -c --host={host_string} --username={db_user} '
            '--blobs --format=c {db_name} -n public | '
            'pg_restore -U {db_postgresql_postgres_user} --create '
            '--dbname={db_name}')

    @task
    def drop_views(self, name=None, site=None):
        """
        Drops all views.
        """
        raise NotImplementedError
    #        SELECT 'DROP VIEW ' || table_name || ';'
    #        FROM information_schema.views
    #        WHERE table_schema NOT IN ('pg_catalog', 'information_schema')
    #        AND table_name !~ '^pg_';
            # http://stackoverflow.com/questions/13643831/drop-all-views-postgresql
    #        DO$$
    #        BEGIN
    #
    #        EXECUTE (
    #           SELECT string_agg('DROP VIEW ' || t.oid::regclass || ';', ' ')  -- CASCADE?
    #           FROM   pg_class t
    #           JOIN   pg_namespace n ON n.oid = t.relnamespace
    #           WHERE  t.relkind = 'v'
    #           AND    n.nspname = 'my_messed_up_schema'
    #           );
    #
    #        END
    #        $$

    @task
    def exists(self, name='default', site=None, use_root=False):
        """
        Returns true if a database with the given name exists. False otherwise.
        """

        r = self.database_renderer(name=name, site=site)

        if int(use_root):
            kwargs = dict(
                db_user=r.env.get('db_root_username', 'postgres'),
                db_password=r.env.get('db_root_password', 'password'),
                db_host=r.env.db_host,
                db_name=r.env.db_name,
            )
            r.env.update(kwargs)

        # Set .pgpass file.
        if r.env.db_password:
            self.write_pgpass(name=name, root=use_root)

        ret = r.run(
            'psql --username={db_user} --host={db_host} -l | grep -v test_ | grep {db_name} | wc -l', ignore_errors=True
        )
        if ret is not None:
            if 'password authentication failed' in ret:
                ret = False
            else:
                ret = int(ret.split('\n')[-1]) >= 1

            print('%s database on site %s %s exist' % (r.env.db_name, self.genv.SITE, 'DOES' if ret else 'DOES NOT'))

        return ret

    @task
    def execute(
        self, sql, name='default', site=None, as_db_root_user=False, ignore_errors=False, no_db=False, no_pager=False,
        timeout=None, **kwargs
    ):
        """
        Run SQL command with psql.

        Args:
            sql (str): Command to run
            name (str): Database name
            site (str): Site name
            as_db_root_user (int/bool): If truthy, run as db_root_username; otherwise, run as db_user
            ignore_errors (int/bool): If truthy, wrap task in settings(warn_only=True)
            no_db (int/bool): If truthy, omit --dbname argument (as when creating a database)
            no_pager (int/bool): If truthy, disable psql pager, to prevent interactive less-style prompt.
            timeout (int/None): If not None, terminate command after this many seconds.
        """
        r = self.database_renderer(name=name, site=site)

        sql = sql.strip()
        if not sql.endswith(';'):
            sql += ';' # Terminate SQL statements with semicolon
        r.env.sql = sql
        r.env.dbname_arg = '' if no_db else f'--dbname {r.env.db_name}'
        r.env.pager_arg = '-P pager=off' if int(no_pager) else ''

        if as_db_root_user and r.env.db_host in {'localhost', '127.0.0.1'}:
            # Run locally as db root user with sudo -U, relying on pg_hba.conf or other Postgres auth
            ret = r.sudo(
                'psql --user={db_root_username} --no-password --host={db_host} {dbname_arg} {pager_arg} --command="{sql}"',
                user=r.env.postgres_user, ignore_errors=ignore_errors, timeout=timeout)
        elif as_db_root_user:
            # Run on remote database as db root user, relying on .pgpass or other Postgres auth
            ret = r.run(
                'psql --user={db_root_username} --no-password --host={db_host} {dbname_arg} {pager_arg} --command="{sql}"',
                ignore_errors=ignore_errors, timeout=timeout)
        else:
            # Run as normal db user, relying on .pgpass or other Postgres auth
            ret = r.run(
                'psql --user={db_user} --no-password --host={db_host} {dbname_arg} {pager_arg} --command="{sql}"',
                ignore_errors=ignore_errors, timeout=timeout)

        return ret

    @task
    def execute_file(self, filename, name='default', site=None, **kwargs):
        """
        Run SQL file with psql.
        """
        r = self.database_renderer(name=name, site=site)
        r.env.filename = filename
        #r.run('psql --user={postgres_user} --no-password -d {db_name} -a -f {filename}')
        r.run('psql --user={db_user} --no-password --host={db_host} -d {db_name} -a -f {filename}')

    @task
    def create(self, name='default', site=None, **kwargs):
        r = self.database_renderer(name=name, site=site)

        # Load site-specific satchel settings.
        site = site or self.genv.SITE
        self.set_site_specifics(site)

        # Create role/user.
        r.pc('Creating user...')
        self.execute(
            "CREATE USER {db_user} WITH PASSWORD '{db_password}';",
            name=name, site=site, as_db_root_user=True, ignore_errors=True, no_db=True)
        # Grant user role to root role (prevents "must be member of role <db_user>" errors in RDS)
        self.execute("GRANT {db_user} TO {db_root_username};", name=name, site=site, as_db_root_user=True, no_db=True)

        # Create db
        r.pc('Creating database...')
        ret = self.execute(
            "CREATE DATABASE {db_name} WITH OWNER={db_user} ENCODING='{encoding}' LC_CTYPE='{locale}' LC_COLLATE='{locale}';",
            name=name, site=site, as_db_root_user=True, ignore_errors=True, no_db=True)
        if isinstance(ret, six.string_types) and 'ERROR:' in ret and 'already exists' not in ret:
            raise Exception('Error creating database: %s' % ret)

        if r.env.schema_mt:
            # Create schema
            # This assumes each schema is associated with a unique user.
            r.pc('Creating schema...')
            self.execute(
                "CREATE SCHEMA IF NOT EXISTS {db_schema}; "
                "GRANT ALL PRIVILEGES ON SCHEMA {db_schema} to {db_user}; "
                "ALTER ROLE {db_user} SET search_path TO {db_schema};", name=name, site=site, as_db_root_user=True)

        r.pc('Enabling plpgsql on database...')
        r.run('createlang -U postgres plpgsql {db_name} || true', ignore_errors=True)

    @task
    def drop_connections(self, name=None, site=None, drop_other_usenames=0):
        """
        Drop non-root connections to the target site database.

        Useful for quickly killing deadlocks, or preventing interference with a db snapshot load.
        Uses pg_terminate_backend() to send SIGTERM to all running non-shell (psql) query processes.
        Times out after 10 seconds, since terminating connections is expected to be a very fast operation.

        Args:
            name (str): db name
            site (str): site slug
            drop_other_usenames (int): If 1, also drop connections from other non-root usenames (sites).
        """
        site = site or self.genv.SITE
        drop_other_usenames = int(drop_other_usenames)

        print('Dropping database connections...')

        r = self.database_renderer(name=name, site=site)
        external_ip = (r.run('wget http://ipecho.net/plain -O - -q ; echo') or '').strip()
        if r.env.db_root_username == 'postgres' and r.env.db_host == external_ip:
            r.env.db_host = '127.0.0.1'

        if drop_other_usenames:
            r.env.usename_condition = "usename != '{db_root_username}'"
        else:
            r.env.usename_condition = "usename = '{SITE}'"

        self.execute(
            "SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = '{db_name}' "
            "AND {usename_condition} AND application_name != 'psql';",
            name=name, site=site, as_db_root_user=True, ignore_errors=True, timeout=10)

    @task
    #@runs_once Interferes with global methods that want to load multiple databases.
    def load(self, dump_fn='', prep_only=0, force_upload=0, from_local=0, name=None, site=None, dest_dir=None, force_host=None):
        """
        Restore a database snapshot onto the target database server.

        Args:
            prep_only (int): If 1, commands for preparing the load will be generated, but not the command to finally load
                the snapshot.
            force_host (str): Force loading on the given host.
        """

        site = site or self.genv.SITE

        r = self.database_renderer(name=name, site=site)

        # Load site-specific satchel settings.
        try:
            self.set_site_specifics(site)
        except KeyError:
            # Sometimes databases have a logical site associated with them but no explicit site settings.
            pass

        # Render the snapshot filename.
        r.env.dump_fn = self.get_default_db_fn(fn_template=dump_fn, dest_dir=dest_dir)

        prep_only = int(prep_only)

        missing_local_dump_error = r.format("Database dump file {dump_fn} does not exist.")

        # Copy snapshot file to target.
        if self.is_local:
            r.env.remote_dump_fn = dump_fn
        else:
            r.env.remote_dump_fn = '/tmp/' + os.path.split(r.env.dump_fn)[-1]

        if not prep_only and not self.is_local:
            if not self.dryrun:
                assert os.path.isfile(r.env.dump_fn), missing_local_dump_error
            self.upload_snapshot(name=name, site=site, local_dump_fn=r.env.dump_fn, remote_dump_fn=r.env.remote_dump_fn)

        if self.is_local and not prep_only and not self.dryrun:
            assert os.path.isfile(r.env.dump_fn), missing_local_dump_error

        if force_host:
            r.env.db_host = force_host

        # Drop/create schema/db
        if r.env.schema_mt:
            # db may already exist on multitenant sites.
            # Use ignore_errors in place of missing "IF EXISTS" syntax for CREATE DATABASE.
            self.execute("CREATE DATABASE {db_name};", name=name, site=site, as_db_root_user=True, ignore_errors=True, no_db=True)
            # Disconnect all other users so we can drop the schema if it exists.
            self.drop_connections(name=name, site=site)
            self.execute("DROP SCHEMA IF EXISTS {db_schema} CASCADE;", name=name, site=site, as_db_root_user=True)
        else:
            # Disconnect all other users so we can drop the database if it exists.
            self.drop_connections(name=name, site=site, drop_other_usenames=1)
            self.execute("DROP DATABASE IF EXISTS {db_name};", name=name, site=site, as_db_root_user=True, no_db=True)
            self.execute("CREATE DATABASE {db_name};", name=name, site=site, as_db_root_user=True, no_db=True)

        # Create PostGIS extensions
        if r.env.engine == POSTGIS:
            self.execute("CREATE EXTENSION postgis;", name=name, site=site, as_db_root_user=True, ignore_errors=True)
            self.execute("CREATE EXTENSION postgis_topology;", name=name, site=site, as_db_root_user=True, ignore_errors=True)

        # Reassign user-owned objects to root user and drop user schemas
        # (user may not exist yet, so ignore errors)
        if not r.env.schema_mt:
            self.execute(
                "REASSIGN OWNED BY {db_user} TO {db_root_username}; "
                "DROP OWNED BY {db_user} CASCADE;", name=name, site=site, as_db_root_user=True, ignore_errors=True)

        # Create db user and assign privileges as appropriate
        self.execute(
            "DROP USER IF EXISTS {db_user}; "
            "CREATE USER {db_user} WITH PASSWORD '{db_password}';", name=name, site=site, as_db_root_user=True)
        if not r.env.schema_mt:
            self.execute("GRANT ALL PRIVILEGES ON DATABASE {db_name} to {db_user};", name=name, site=site, as_db_root_user=True)
            for createlang in r.env.createlangs:
                r.env.createlang = createlang
                r.sudo('createlang -U {db_root_username} --host={db_host} {createlang} {db_name} || true', user=r.env.postgres_user)

        if r.env.schema_mt:
            # Create schema and assign to user. This assumes each schema is associated with a unique user.
            self.execute(
                "CREATE SCHEMA IF NOT EXISTS {db_schema}; "
                "GRANT ALL PRIVILEGES ON SCHEMA {db_schema} to {db_user}; "
                "ALTER ROLE {db_user} SET search_path TO {db_schema};", name=name, site=site, as_db_root_user=True)

        if not prep_only:
            # Run load command (as postgres user, if database is hosted locally)
            if r.env.db_host in {'localhost', '127.0.0.1'}:
                r.sudo(r.env.load_command, user=r.env.postgres_user)
            else:
                r.run(r.env.load_command)

    @task
    def shell(self, name='default', site=None, **kwargs):
        """
        Open a SQL shell to the given database, assuming the configured database and user support this feature.
        """
        r = self.database_renderer(name=name, site=site)
        self.write_pgpass(name=name, site=site, root=True)

        db_name = kwargs.get('db_name')
        if db_name:
            r.env.db_name = db_name
            r.run('/bin/bash -i -c "psql --username={db_root_username} --host={db_host} --dbname={db_name}"')
        else:
            r.run('/bin/bash -i -c "psql --username={db_root_username} --host={db_host}"')

    @task
    def configure_apt_repository(self):
        r = self.local_renderer
        r.sudo("add-apt-repository '{apt_repo}'")
        r.sudo('wget --quiet -O - {apt_key} | apt-key add -')
        r.sudo('apt-get update')

    @task
    def drop_database(self, name):
        """
        Delete a PostgreSQL database.

        Example::

            import burlap

            # Remove DB if it exists
            if burlap.postgres.database_exists('myapp'):
                burlap.postgres.drop_database('myapp')

        """
        self.sudo('dropdb %s' % (name,), user='postgres', ignore_errors=True)

    @task
    def version(self):
        r = self.local_renderer
        if self.dryrun:
            v = r.env.version_command
        else:
            v = (r.run('echo {version_command}') or '').strip()
        self.vprint('postgresql version:', v)

        # Postgres 10+ doesn't use minor version in /etc/postgresql/ subdirectory names
        if isinstance(v, str) and not v.startswith('9'):
            v = v.split('.')[0]

        self.vprint('postgresql subdirectory version:', v)
        return v

    @task
    def load_table(self, table_name, src, dst='localhost', name=None, site=None):
        """
        Directly transfers a table between two databases.
        """
        #TODO: incomplete
        r = self.database_renderer(name=name, site=site)
        r.env.table_name = table_name
        r.run('psql --user={dst_db_user} --host={dst_db_host} --command="DROP TABLE IF EXISTS {table_name} CASCADE;"')
        r.run('pg_dump -t {table_name} --user={dst_db_user} --host={dst_db_host} | psql --user={src_db_user} --host={src_db_host}')

    @task
    def write_pg_hba_conf(self):
        r = self.local_renderer
        if 'pg_version' not in r.env:
            r.env.pg_version = self.version()# or r.env.default_version
        r.pc('Writing pg_hba.conf...')
        r.sudo('cp /etc/postgresql/{pg_version}/main/pg_hba.conf /etc/postgresql/{pg_version}/main/pg_hba.conf.$(date +%Y%m%d%H%M).bak')
        fn = self.render_to_file('postgresql/pg_hba.template.conf')
        r.put(
            local_path=fn,
            remote_path='/etc/postgresql/{pg_version}/main/pg_hba.conf',
            use_sudo=True,
        )

    @task(precursors=['packager', 'user', 'locales'])
    def configure(self, *args, **kwargs):
        #TODO:set postgres user password?
        #https://help.ubuntu.com/community/PostgreSQL
        #set postgres ident in pg_hba.conf
        #sudo -u postgres psql postgres
        #sudo service postgresql restart
        #sudo -u postgres psql
        #\password postgres
        r = self.local_renderer

        #self.install_packages()

        if r.env.apt_repo_enabled:
            self.configure_apt_repository()

        r.env.pg_version = self.version()# or r.env.default_version

        self.write_pg_hba_conf()

#         r.pc('Backing up PostgreSQL configuration files...')
        r.sudo('cp /etc/postgresql/{pg_version}/main/postgresql.conf /etc/postgresql/{pg_version}/main/postgresql.conf.$(date +%Y%m%d%H%M).bak')
        r.pc('Enabling auto-vacuuming...')
        r.sed(
            filename='/etc/postgresql/{pg_version}/main/postgresql.conf',
            before='#autovacuum = on',
            after='autovacuum = on',
            backup='',
            use_sudo=True,
        )
        r.sed(
            filename='/etc/postgresql/{pg_version}/main/postgresql.conf',
            before='#track_counts = on',
            after='track_counts = on',
            backup='',
            use_sudo=True,
        )

        r.sudo('service postgresql restart')


class PostgreSQLClientSatchel(Satchel):

    name = 'postgresqlclient'

    def set_defaults(self):
        self.env.force_version = None

    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['postgresql-client'],
            UBUNTU: ['postgresql-client-11'],
            (UBUNTU, '12.04'): ['postgresql-client-9.1'],
            (UBUNTU, '14.04'): ['postgresql-client-9.3'],
            (UBUNTU, '16.04'): ['postgresql-client-10'],
            (UBUNTU, '18.04'): ['postgresql-client-11'],
            (UBUNTU, '20.04'): ['postgresql-client-13'],
        }

    #https://askubuntu.com/questions/831292/how-to-install-postgresql-9-6-on-any-ubuntu-version
    @property
    def packager_repositories(self):
        ver = self.os_version
        if ver.type == LINUX:
            if ver.distro == UBUNTU:
                if ver.release == '16.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ xenial-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
                if ver.release == '18.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ bionic-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
                if ver.release == '20.04':
                    return {
                        APT: ['deb http://apt.postgresql.org/pub/repos/apt/ focal-pgdg main'],
                        APT_KEY: ['https://www.postgresql.org/media/keys/ACCC4CF8.asc',],
                    }
        raise NotImplementedError

    @task(precursors=['packager'])
    def configure(self, *args, **kwargs):
        pass

postgresql = PostgreSQLSatchel()
PostgreSQLClientSatchel()

write_postgres_pgpass = postgresql.write_pgpass



import os

from fabric.api import settings

from burlap.constants import *
from burlap import ServiceSatchel
from burlap.decorators import task
from burlap.trackers import FilesystemTracker, SettingsTracker, ORTracker

class BuildBotSatchel(ServiceSatchel):
    """
    Configures a Buildbot master and worker on the first host,
    with a worker on every additional host.
    """

    name = 'buildbot'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['git'],
        }

    def set_defaults(self):

        self.env.project_dir = '/usr/local/myproject'
        self.env.virtualenv_dir = '/usr/local/myproject/.env'
        self.env.home_dir = '/var/lib/{bb_user}'
        self.env.src_dir = 'src/buildbot'

        self.env.ssh_bin = '{home_dir}/bin'
        self.env.ssh_dir = '{home_dir}/.ssh'
        self.env.ssh_private_key = None # should be a *.pem file
        self.env.ssh_public_key = None # should be a *.pub file

        # Must match the main user, or otherwise we get rsync errors.
        self.env.user = 'ubuntu'
        self.env.group = 'ubuntu'

        self.env.bb_user = 'buildbot'
        self.env.bb_group = 'buildbot'

        self.env.manhole_user = 'admin'
        self.env.manhole_port = 1234

        self.env.perms = '777'

        self.env.extra_deploy_paths = ['buildbot/worker/info/', 'buildbot/worker/buildbot.tac']

        self.env.delete_deploy_paths = []

        self.env.check_ok = True
        self.env.check_ok_paths = {} # {branch: (url, text)}

        self.env.rsync_paths = [] # [[from_path, to_path, user, group]]

        self.env.enable_apache_site = True

        self.env.requirements = 'pip-requirements.txt'

        self.env.use_ssh_key = False

        self.env.pid_path = 'buildbot/{type}/twistd.pid'

        self.env.rsync_excludes = []

        self.env.worker_names = ['worker']

        #DEPRECATED
        self.env.cron_check_enabled = False
        self.env.cron_check_schedule = '0,30 * * * *'
        self.env.cron_check_user = 'root'
        self.env.cron_cron_check_worker_pid_path = None
        self.env.cron_check_command_template = 'buildbot/check_buildbot.sh.template'
        self.env.cron_check_command_path = '/usr/local/bin/check_buildbot.sh'
        self.env.cron_check_crontab_template = 'buildbot/etc_crond_buildbot.template'
        self.env.cron_check_crontab_path = '/etc/cron.d/buildbot'

        self.env.supervisor_enabled = False

        self.env.service_commands = {
#             START:{
#                 UBUNTU: 'service apache2 start',
#             },
#             STOP:{
#                 UBUNTU: 'service apache2 stop',
#             },
#             DISABLE:{
#                 (UBUNTU, '14.04'): 'update-rc.d -f apache2 remove',
#             },
#             ENABLE:{
#                 (UBUNTU, '14.04'): 'update-rc.d apache2 defaults',
#             },
#             RELOAD:{
#                 UBUNTU: 'service apache2 reload',
#             },
#             RESTART:{
#                 # Note, the sleep 5 is necessary because the stop/start appears to
#                 # happen in the background but gets aborted if Fabric exits before
#                 # it completes.
#                 UBUNTU: 'service apache2 restart; sleep 3',
#             },
        }

    @property
    def is_first_host(self):
        return self.genv.hosts[0] == self.genv.host_string

    @task
    def restart(self):
        r = self.local_renderer
        if r.env.supervisor_enabled:
            r.sudo('supervisorctl restart all')
        else:
            self.restart_master(ignore_errors=True)
            self.restart_worker(ignore_errors=True)

    @property
    def restart_master_command(self):
        r = self.local_renderer
        r.env.restart_master_command = r.format(
            'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
            '{virtualenv_dir}/bin/buildbot restart master"')
        return r.env.restart_master_command

    @task
    def restart_master(self, ignore_errors=None):
        if not self.is_first_host:
            return
        ignore_errors = self.ignore_errors if ignore_errors is None else ignore_errors
        r = self.local_renderer
        s = {'warn_only':True} if ignore_errors else {}
        with settings(**s):
            r.run(self.restart_master_command)

    @task
    def upgrade_master(self):
        r = self.local_renderer
        r.run('sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; {virtualenv_dir}/bin/buildbot upgrade-master {project_dir}/src/buildbot/master"')

    def get_restart_worker_command(self, name=None):
        r = self.local_renderer
        parts = []
        for worker_name in self.get_worker_names_for_current_host():
            if name and worker_name != name:
                continue
            r.env.worker_name = worker_name
            parts.append('cd {project_dir}/src/buildbot; {virtualenv_dir}/bin/buildbot-worker restart %s;' % worker_name)
        if not parts:
            return
        parts = ' '.join(parts)
        r.env.restart_worker_command = r.format('sudo -u {bb_user} bash -c "%s"' % parts)
        return r.env.restart_worker_command

    @task
    def restart_worker(self, ignore_errors=None, name=None):
        ignore_errors = self.ignore_errors if ignore_errors is None else ignore_errors
        r = self.local_renderer
        s = {'warn_only':True} if ignore_errors else {}
        with settings(**s):
            cmd = self.get_restart_worker_command(name=name)
            if cmd:
                print('cmd:', cmd)
                r.run(cmd)

    @task
    def start(self):
        r = self.local_renderer
        s = {'warn_only':True} if self.ignore_errors else {}
        with settings(**s):
            if r.env.supervisor_enabled:
                r.sudo('supervisorctl start all')
            else:
                r.run(
                    'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                    '{virtualenv_dir}/bin/buildbot start master"')
                for worker_name in self.get_worker_names_for_current_host():
                    r.env.worker_name = worker_name
                    r.run(
                        'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                        '{virtualenv_dir}/bin/buildbot-worker start {worker_name}"')

    @property
    def host_index(self):
        return self.genv.hosts.index(self.genv.host_string)

    def get_worker_names_for_current_host(self):
        r = self.local_renderer
        names = []
        host_index = self.host_index
        for i, worker_name in enumerate(r.env.worker_names):
            if i == host_index:
                names.append(worker_name)
        return names

    @task
    def stop(self):
        r = self.local_renderer
        s = {'warn_only': True}
        with settings(**s):
            if r.env.supervisor_enabled:
                r.sudo('supervisorctl stop all')
            else:
                if self.is_first_host:
                    r.run(
                        'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                        '{virtualenv_dir}/bin/buildbot stop master"')
                for worker_name in self.get_worker_names_for_current_host():
                    r.env.worker_name = worker_name
                    with settings(warn_only=True):
                        r.run(
                            'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                            '{virtualenv_dir}/bin/buildbot-worker stop {worker_name}"')

    @task
    def reload(self):
        r = self.local_renderer
        s = {'warn_only':True} if self.ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "'
                'cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot reconfig master"')

    @task
    def set_permissions(self):
        """
        Assign ownership of project and home directories to Buildbot user.
        """
        r = self.local_renderer
        r.sudo('chown -R {bb_user}:{bb_group} {home_dir}')
        r.sudo('chown -R {bb_user}:{bb_group} {project_dir}')
        r.sudo('chmod -R a+rw {project_dir}') # Allow access by other users (e.g. www-data?)

    @task
    def rsync_paths(self):
        """
        Transfer miscellaneous extra files to Buildbot host.
        """
        r = self.local_renderer
        for from_path, to_path, to_user, to_group, to_perms in r.env.rsync_paths:
            assert os.path.isfile(from_path), 'File %s does not exist locally.' % from_path
            r.env.from_path = from_path
            r.env.to_path = to_path
            r.env.to_user = to_user
            r.env.to_group = to_group
            r.env.to_path_fq = os.path.join(to_path, os.path.split(from_path)[-1])
            r.env.to_perms = to_perms
            r.put(local_path=r.env.from_path, remote_path=r.env.to_path, use_sudo=True)
            r.sudo('chown {to_user}:{to_group} {to_path_fq}')
            r.sudo('chmod {to_perms} {to_path_fq}')

    @task
    def setup_dir(self, clean=0):
        pip = self.get_satchel('pip')
        clean = int(clean)
        r = self.local_renderer
        if clean:
            r.sudo('rm -Rf {virtualenv_dir} || true')
        r.sudo('mkdir -p {project_dir}')
        pip.update_install(
            requirements=self.env.requirements,
            virtualenv_dir=self.env.virtualenv_dir,
            user=self.env.user,
            group=self.env.group,
            perms=self.env.perms,
        )
        #r.sudo('{virtualenv_dir}/bin/pip install -U pip')
        r.sudo('chown -R {user}:{group} {project_dir}')
        r.sudo('chmod -R {perms} {project_dir}')

    @task
    def deploy_code(self):
        r = self.local_renderer

        # Temporarily assign ownership of project directory to deploy user to facilitate rsync transfers.
        # Will be reverted to Buildbot user later with set_permissions().
        r.sudo('chown -R {user}:{group} {project_dir}')

        # Remove unnecessary files.
        for delete_path in r.env.delete_deploy_paths:
            r.sudo('rm -Rf %s' % delete_path)

        exclude_str = ' '.join(['--exclude=%s' % path for path in r.env.rsync_excludes]) + ' '

        # Update project files with rsync.
        r.local('rsync '
            '--recursive --verbose --perms --times --links '
            '--compress --copy-links '
            '--exclude=*.pyc --exclude=gitpoller-workdir --exclude=*.log --exclude=twistd.pid '
            '--exclude=*.sqlite '
            '--exclude=build '
            '--exclude=worker '
            '--exclude=*_runtests ' + \
            exclude_str + \
            '--delete --delete-before '
            '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}" '
            'src {user}@{host_string}:{project_dir}')

        # Sync extra deploy paths.
        for path in self.env.extra_deploy_paths:
            if path.endswith('/'):
                # Directory.
                r.env.tmp_path = path[:-1]
                r.env.tmp_remote_path = os.path.split(path[:-1])[0]
                r.sudo('mkdir -p {project_dir}/src/{tmp_remote_path}')
            else:
                # File.
                r.env.tmp_path = path
                r.env.tmp_remote_path = os.path.split(path)[0]

            r.pc('Fixing permissions for {}...'.format(path))
            r.local('rsync '
                '--recursive --verbose --times --omit-dir-times --links '
                '--compress --copy-links '
                '--delete --delete-before '
                '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}" '
                'src/{tmp_path} {user}@{host_string}:{project_dir}/src/{tmp_remote_path}')

        self.rsync_paths()
        self.set_permissions()

    @task
    def view_log(self, name='master', section='twistd'):
        r = self.local_renderer
        r.env.name = name
        assert section in ('twistd', 'http')
        r.env.section = section
        r.run('tail -f --lines=100 {project_dir}/src/buildbot/{name}/{section}.log')

    @task
    def manhole(self):
        r = self.local_renderer
        r.env.host_string = '127.0.0.1'#self.genv.host_string
        r.run('ssh -p{manhole_port} {manhole_user}@{host_string}')

    @task
    def check_ok(self):
        """
        Ensures all tests have passed for this branch.

        This should be called before deployment, to prevent accidental deployment of code
        that hasn't passed automated testing.
        """
        import requests

        if not self.env.check_ok:
            return

        # Find current git branch.
        branch_name = self._local('git rev-parse --abbrev-ref HEAD', capture=True).strip()

        check_ok_paths = self.env.check_ok_paths or {}

        if branch_name in check_ok_paths:
            check = check_ok_paths[branch_name]
            if 'username' in check:
                auth = (check['username'], check['password'])
            else:
                auth = None
            ret = requests.get(check['url'], auth=auth)
            passed = check['text'] in ret.content
            assert passed, 'Check failed: %s' % check['url']

    @task
    def setup_apache(self):
        r = self.local_renderer
        if r.env.enable_apache_site and self.is_first_host:
            apache = self.get_satchel('apache')
            apache.enable_mod('proxy_http')

    @task
    def setup_user(self):
        r = self.local_renderer
        user = self.get_satchel('user')
        group = self.get_satchel('group')
        group.create(self.env.bb_group)
        user.create(
            username=r.env.bb_user,
            groups=r.env.bb_group,
            create_home=True,
            home_dir=r.format(r.env.home_dir),
            password=False,
        )
        #user.create(self.env.bb_user, self.env.bb_group)

    def deploy_pre_run(self):
        self.check_ok()

    @task
    def configure_ssh_key(self):
        r = self.local_renderer
        r.env.private_remote_path = '{ssh_dir}/id_rsa'
        r.env.public_remote_path = '{ssh_dir}/id_rsa.pub'
        if r.env.use_ssh_key:
            #https://www.cyberciti.biz/faq/how-to-set-up-ssh-keys-on-linux-unix/
            assert r.env.ssh_private_key, 'No SSH private key specified!'
            assert r.env.ssh_public_key, 'No SSH public key specified!'
            r.sudo('mkdir -p {ssh_dir}')
            r.put(local_path=r.env.ssh_private_key, remote_path=r.env.private_remote_path, use_sudo=True)
            r.put(local_path=r.env.ssh_public_key, remote_path=r.env.public_remote_path, use_sudo=True)
            r.sudo('chown -R {bb_user}:{bb_group} {ssh_dir}')
            r.sudo('chmod -R 0770 {ssh_dir}')
            r.sudo('chmod -R 0600 {ssh_dir}/*') # SSH keys must use restrictive permissions.
        else:
            r.sudo('rm -F {private_remote_path}')
            r.sudo('rm -F {public_remote_path}')

    @task
    def delete_logs(self):
        r = self.local_renderer
        names = ('master', 'worker')
        logs = ('twistd.log', 'http.log')
        for name in names:
            r.env.name = name
            for log in logs:
                r.env.log = log
                r.sudo('rm -f {project_dir}/src/buildbot/{name}/{log}')

    @task
    def install_cron_check(self):
        r = self.local_renderer

        # Install script to perform the actual check.
        assert r.env.cron_check_worker_pid_path, 'Worker PID path not set.'
        self.restart_master_command
        self.get_restart_worker_command()
        self.install_script(
            local_path=r.env.cron_check_command_template,
            remote_path=r.env.cron_check_command_path,
            render=True,
            extra=r.collect_genv())
        r.sudo('chown root:root {cron_check_command_path}')

        # Install crontab to schedule running the script.
        self.install_script(
            local_path=r.env.cron_check_crontab_template,
            remote_path=r.env.cron_check_crontab_path,
            render=True,
            extra=r.collect_genv())
        r.sudo('chown root:root {cron_check_crontab_path}')
        r.sudo('chmod 600 {cron_check_crontab_path}')
        r.sudo('service cron restart')

    @task
    def uninstall_cron_check(self):
        r = self.local_renderer
        r.sudo('rm -f {cron_check_crontab_path}')
        r.sudo('rm -f {cron_check_command_path}')
        r.sudo('service cron restart')

    @task
    def install_supervisor(self):
        r = self.local_renderer
        if r.env.supervisor_enabled:
            extra = {'workerN': r.genv.hosts.index(r.genv.host_string)+1}
            self.install_script(
                local_path='buildbot/supervisor.conf.template',
                remote_path='/etc/supervisor/conf.d/buildbot.conf',
                render=True,
                extra=extra)
        else:
            r.sudo('rm -f /etc/supervisor/conf.d/buildbot.conf')
            r.sudo('supervisorctl reload')

    # @task
    # def update_cron_check(self):
        # if not self.is_first_host:
            # return
        # elif self.param_changed_to('cron_check_enabled', True):
            # self.install_cron_check()
        # elif self.param_changed_to('cron_check_enabled', False):
            # self.uninstall_cron_check()

    @task
    def uninstall(self):
        r = self.local_renderer

        with settings(warn_only=True):
            self.stop()

        self.uninstall_cron_check()

        r.sudo('rm -Rf {virtualenv_dir} || true')
        r.sudo('rm -Rf {project_dir} || true')

    def get_trackers(self):
        r = self.local_renderer
        return [

            SettingsTracker(
                satchel=self,
                names='bb_user bb_group home_dir',
                action=self.setup_user),

            ORTracker(
                SettingsTracker(satchel=self, names='virtualenv_dir project_dir'),
                FilesystemTracker(base_dir=r.format('roles/{ROLE}'), extensions='pip-requirements.txt'),
                action=self.setup_dir),

            SettingsTracker(
                satchel=self,
                names='enable_apache_site',
                action=self.setup_apache),

            SettingsTracker(
                satchel=self,
                names='ssh_dir ssh_private_key ssh_public_key bb_user bb_group',
                action=self.configure_ssh_key),

            FilesystemTracker(
                base_dir=r.format(r.env.src_dir), extensions='*.py *.tac *.cfg htpasswd',
                action=self.deploy_code),

            # SettingsTracker(
                # satchel=self,
                # names='project_dir cron_path cron_perms',
                # action=self.install_cron),

            SettingsTracker(
                satchel=self,
                # names='project_dir cron_path cron_perms',
                action=self.install_supervisor),

            # SettingsTracker(
                # satchel=self,
                # names='cron_check_enabled',
                # action=self.update_cron_check),

        ]

    @task(precursors=['packager', 'user', 'apache'])
    def configure(self):
        r = self.local_renderer

        has_changes = self.has_changes
        if has_changes:
            self.vprint('Stopping any existing buildbot server...')
            with settings(warn_only=True):
                self.stop()

        super().configure()

        if has_changes:
            self.restart()
            if self.is_first_host:
                apache = self.get_satchel('apache')
                apache.reload()

buildbot = BuildBotSatchel()



"""
Run specific tests like:

    tox -c tox-full.ini -e py37-ubuntu_18_04_64 -- -s burlap/tests/functional_tests/test_selenium.py::SeleniumTests::test_selenium

"""
from unittest import skip

from fabric.contrib.files import exists

from burlap.common import set_verbose
from burlap.selenium import selenium
from burlap.tests.functional_tests.base import TestCase
from burlap.deploy import deploy as deploy_satchel

class SeleniumTests(TestCase):

    def test_paths(self):
        print('deploy_satchel.env.lockfile_path:', deploy_satchel.env.lockfile_path)
        print('deploy_satchel.env.data_dir:', deploy_satchel.env.data_dir)
        assert deploy_satchel.env.lockfile_path == '/tmp/burlap_unittests/deploy.lock'
        assert deploy_satchel.env.data_dir == '/tmp/burlap_unittests'
        assert deploy_satchel.manifest_filename == '/tmp/burlap_unittests/manifest.yaml'

    @skip('dns failure')
    def test_selenium(self):
        try:
            print('deploy_satchel.env.lockfile_path:', deploy_satchel.env.lockfile_path)
            print('deploy_satchel.env.data_dir:', deploy_satchel.env.data_dir)
            assert deploy_satchel.env.lockfile_path == '/tmp/burlap_unittests/deploy.lock'
            assert deploy_satchel.env.data_dir == '/tmp/burlap_unittests'
            assert deploy_satchel.manifest_filename == '/tmp/burlap_unittests/manifest.yaml'

            set_verbose(True)
            print('selenium.geckodriver_path:', selenium.geckodriver_path)
            selenium.genv.ROLE = 'local'
            selenium.genv.services = ['selenium']
            selenium.clear_caches()

            print('Enabling selenium/gecko to install and track old version.')
            print('selenium._last_manifest.1:', selenium._last_manifest)
            print('selenium.last_manifest.1:', selenium.last_manifest)
            selenium.env.enabled = True
            selenium.env.geckodriver_version = '0.13.0'
            selenium.clear_local_renderer()
            assert selenium.get_target_geckodriver_version_number() == '0.13.0'
            print('Configuring selenium...')
            selenium.configure()
            print('selenium._last_manifest.2:', selenium._last_manifest)
            print('selenium.last_manifest.2:', selenium.last_manifest)
            print('Writing manifest...')
            deploy_satchel.fake(components=selenium.name)
            deploy_satchel.run('ls -lah %s' % deploy_satchel.manifest_filename)
            deploy_satchel.run('cat %s' % deploy_satchel.manifest_filename)
            print('selenium._last_manifest.3:', selenium._last_manifest)
            print('selenium.last_manifest.3:', selenium.last_manifest)

            print('Confirming install succeeded...')
            assert exists(selenium.geckodriver_path)
            assert not selenium.check_for_change()
            output = selenium.run('geckodriver --version')
            print('Geckodriver version:', output)
            expected_version = selenium.env.geckodriver_version
            assert expected_version in output

            print('Updating configuration to track the most recent version...')
            selenium.env.geckodriver_version = None
            selenium.clear_local_renderer()
            assert selenium.get_target_geckodriver_version_number() != '0.13.0'
            assert selenium.last_manifest.fingerprint == '0.13.0'

            print('Confirm we now see a pending change...')
            assert selenium.check_for_change()

            print('-'*80)
            print('Applying change...')
            selenium.configure()
            deploy_satchel.purge()
            print('-'*80)
            print('Thumbprinting...')
            deploy_satchel.fake(components=selenium.name)
            print('-'*80)

            print('Confirming the most recent version was installed...')
            expected_version = selenium.get_most_recent_version()
            selenium.clear_caches()
            assert selenium.last_manifest.fingerprint == expected_version
            output = selenium.run('geckodriver --version')
            expected_version = selenium.get_latest_geckodriver_version_number()
            assert expected_version in output
            assert not selenium.check_for_change()

            print('Update configuration to not manage gecko and apply...')
            selenium.env.enabled = False
            selenium.clear_local_renderer()
            selenium.configure()

            # Confirm gecko was uninstalled.
            assert not exists(selenium.geckodriver_path)

        finally:
            selenium.uninstall_geckodriver()

    def test_selenium_2(self):
        print('deploy_satchel.env.lockfile_path:', deploy_satchel.env.lockfile_path)
        print('deploy_satchel.env.data_dir:', deploy_satchel.env.data_dir)
        assert deploy_satchel.env.lockfile_path == '/tmp/burlap_unittests/deploy.lock'
        assert deploy_satchel.env.data_dir == '/tmp/burlap_unittests'
        assert deploy_satchel.manifest_filename == '/tmp/burlap_unittests/manifest.yaml'



from fabric.tasks import WrappedCallableTask as _WrappedCallableTask


class WrappedCallableTask(_WrappedCallableTask):
    """
    A modified version of Fabric's WrappedCallableTask that sets a global
    dryrun variable from a task call.

    The variable is then removed from the function call, allowing all wrapped
    function calls to implicitly use the keyword argument.
    """

    def __init__(self, *args, **kwargs):
        real_module = kwargs.pop('real_module', None)
        super().__init__(*args, **kwargs)
        if real_module:
            self.__module__ = real_module

    def __call__(self, *args, **kwargs):
        from burlap.common import set_dryrun, set_verbose
        if 'dryrun' in kwargs:
            set_dryrun(kwargs['dryrun'])
            del kwargs['dryrun']
        if 'verbose' in kwargs:
            set_verbose(kwargs['verbose'])
            del kwargs['verbose']
        return self.run(*args, **kwargs)

    def run(self, *args, **kwargs):
        from burlap.common import set_dryrun, set_verbose
        if 'dryrun' in kwargs:
            set_dryrun(kwargs['dryrun'])
            del kwargs['dryrun']
        if 'verbose' in kwargs:
            set_verbose(kwargs['verbose'])
            del kwargs['verbose']
        return self.wrapped(*args, **kwargs)




from burlap.files import file # pylint: disable=redefined-builtin
from burlap.user import user
from burlap.tests.functional_tests.base import TestCase

is_dir = file.is_dir

class UserTests(TestCase):

    def test_create_user(self):
        try:
            user.create('user1', create_home=False, password=False)
            assert user.exists('user1')
            assert not is_dir('/home/user1')
        finally:
            user.sudo('userdel -r user1')



from mock import patch

import pytest

from burlap.tests.base import TestCase

class SystemTests(TestCase):

    def test_unsupported_system(self):

        from burlap.system import UnsupportedFamily # pylint: disable=import-outside-toplevel

        with pytest.raises(UnsupportedFamily) as excinfo:

            with patch('burlap.system.distrib_id') as mock_distrib_id:
                mock_distrib_id.return_value = 'foo'
                raise UnsupportedFamily(supported=['debian', 'redhat'])

        exception_msg = str(excinfo.value)
        assert exception_msg == "Unsupported family other (foo). Supported families: debian, redhat"
